create_task_schema = {
    "type": "object",
    "properties": {
        "task_name": {"type": "string", "minLength": 5},
        "args": {"type": "object", "minProperties": 1},
    },
    "required": ["task_name", "args"]
}



from fractalis.data.etlhandler import ETLHandler


class TestHandler(ETLHandler):

    _handler = 'test'

    @staticmethod
    def make_label(descriptor):
        return descriptor.get('label')

    def _get_token_for_credentials(self, server: str, auth: dict) -> str:
        return 'abc'



import os

import pandas as pd

from fractalis.data.etl import ETL


class NumericalETL(ETL):

    name = 'demo_tca_coad_numerical-etl'
    produces = 'numerical_array'

    @staticmethod
    def can_handle(handler: str, descriptor: dict):
        return handler == 'demo_tcga_coad' and \
               descriptor['dataType'] == 'numerical_array'

    def extract(self, server: str, token: str, descriptor: dict):
        path = '{}/data/{}.tsv'.format(
            os.path.dirname(os.path.abspath(__file__)), descriptor['field'])
        raw_data = pd.read_csv(path, sep='\t')
        return raw_data

    def transform(self, raw_data: pd.DataFrame, descriptor: dict):
        df = raw_data
        df[['id']] = df[['id']].astype(str)
        df[['value']] = df[['value']].astype(float)
        df[['feature']] = df[['feature']].astype(str)
        return df



"""This module provides AdaHandler, an implementation of ETLHandler for ADA."""

import logging

import requests

from fractalis.data.etlhandler import ETLHandler


logger = logging.getLogger(__name__)


class AdaHandler(ETLHandler):
    """This ETLHandler provides integration with ADA.

    'Ada provides key infrastructure for secured integration, visualization,
    and analysis of anonymized clinical and experimental data stored in CSV
    and tranSMART format, or provided by RedCAP and Synapse apps.'

    Project URL: https://git-r3lab.uni.lu/peter.banda/ncer-pd
    """

    _handler = 'ada'

    @staticmethod
    def make_label(descriptor: dict) -> str:
        return '{} ({})'.format(descriptor['dictionary']['label'],
                                descriptor['data_set'])

    def _get_token_for_credentials(self, server: str, auth: dict) -> str:
        try:
            user = auth['user']
            passwd = auth['passwd']
            if len(user) == 0 or len(passwd) == 0:
                raise KeyError
        except KeyError as e:
            logger.exception(e)
            raise ValueError("The authentication object must contain the "
                             "non-empty fields 'user' and 'passwd'.")
        r = requests.post(url='{}/login'.format(server),
                          headers={'Accept': 'application/json'},
                          data={'id': user, 'password': passwd},
                          timeout=10)
        if r.status_code != 200:
            error = "Could not authenticate. " \
                    "Reason: [{}]: {}".format(r.status_code, r.text)
            logger.error(error)
            raise ValueError(error)
        cookie = r.headers['Set-Cookie']
        token = [s for s in cookie.split(';')
                 if s.startswith('PLAY2AUTH_SESS_ID')][0]
        token = '='.join(token.split('=')[1:])  # remove PLAY2AUTH_SESS_ID=
        return token



"""This module provides an abstract class for testing whether or not the
ETLs produce valid fractalis standard data."""

import abc
import logging

logger = logging.getLogger(__name__)


class IntegrityCheck(metaclass=abc.ABCMeta):
    """This is an abstract class that provides can be called directly"""

    @property
    @abc.abstractmethod
    def data_type(self) -> str:
        """Specifies a fractalis internal data type."""
        pass

    @classmethod
    def can_handle(cls, data_type: str) -> bool:
        """Test if this checker is responsible for the given data type.
        :param data_type: The data type that the ETL attempts to load.
        :return: True if this checker can handle the type.
        """
        return cls.data_type == data_type

    @staticmethod
    def factory(data_type: str) -> 'IntegrityCheck':
        """A factory that returns a checker object for a given data_type.
        :param data_type: Data type that one wants to test against.
        :return: An instance of IntegrityCheck
        """
        from . import CHECK_REGISTRY
        for Check in CHECK_REGISTRY:
            if Check.can_handle(data_type):
                return Check()
        error = "No IntegrityCheck implementation found " \
                "for data type '{}'".format(data_type)
        logger.error(error)
        raise NotImplementedError(error)

    @abc.abstractmethod
    def check(self, data: object) -> None:
        """Raise if the data have an invalid format. This is okay because
        there is no reason for the ETL to continue otherwise.
        :param data: The data to check.
        """
        pass



"""This module provides the ETLHandler class."""

import os
import abc
import json
import logging
from uuid import uuid4
from typing import List, Union

from fractalis.cleanup import janitor
from fractalis import app, redis, celery
from fractalis.data.etl import ETL


logger = logging.getLogger(__name__)


class ETLHandler(metaclass=abc.ABCMeta):
    """This is an abstract class that provides a factory method to create
    instances of implementations of itself. The main purpose of this class
    is the supervision of all ETL processes belonging to this handler. Besides
    that it takes care of authentication business.
    """

    @property
    @abc.abstractmethod
    def _handler(self) -> str:
        """Used by self.can_handle to check whether the current implementation
        is able to handle the incoming request.
        """
        pass

    def _get_token_for_credentials(self, server: str, auth: dict) -> str:
        """ Authenticate with the server and return a token.
        :param server: The server to authenticate with.
        :param auth: dict containing credentials to auth with API
        :return The token returned by the API.
        """
        raise NotImplementedError()

    def __init__(self, server, auth):
        if not isinstance(server, str) or not server:
            error = ("{} is not a valid server url.".format(server))
            logger.error(error)
            raise ValueError(error)
        if not isinstance(auth, dict):
            error = "'auth' must be a valid dictionary."
            logger.error(error)
            raise ValueError(error)
        if server.endswith('/'):
            server = server[:-1]
        self._server = server
        # if no token is given we have to get one
        try:
            self._token = auth['token']
            if not isinstance(self._token, str) or len(self._token) == 0:
                raise KeyError
        except KeyError:
            logger.info('No token has been provided. '
                        'Attempting to authenticate with the API.')
            try:
                self._token = self._get_token_for_credentials(server, auth)
            except Exception as e:
                logger.exception(e)
                raise ValueError("Could not authenticate with API.")

    @staticmethod
    @abc.abstractmethod
    def make_label(descriptor: dict) -> str:
        """Create a label for the given data descriptor. This label will be
        used to display the loaded data in the front end.
        :param descriptor: Describes the data and is used to download them.
        :return The label (not necessarily unique) to the data."""
        pass

    def create_redis_entry(self, task_id: str, file_path: str,
                           descriptor: dict, data_type: str) -> None:
        """Creates an entry in Redis that contains meta information for the
        data that are to be downloaded.
        :param task_id: Id associated with the loaded data.
        :param file_path: Location of the data on the file system.
        :param descriptor: Describes the data and is used to download them.
        :param data_type: The fractalis internal data type of the loaded data.
        """
        data_state = {
            'task_id': task_id,
            'file_path': file_path,
            'label': self.make_label(descriptor),
            'data_type': data_type,
            'hash': self.descriptor_to_hash(descriptor),
            'meta': {
                'descriptor': descriptor,
            }
        }
        redis.setex(name='data:{}'.format(task_id),
                    value=json.dumps(data_state),
                    time=app.config['FRACTALIS_DATA_LIFETIME'])

    def descriptor_to_hash(self, descriptor: dict) -> int:
        """Compute hash for the given descriptor. Used to identify duplicates.
        :param descriptor: ETL descriptor. Used to identify duplicates.
        :return: Unique hash.
        """
        string = '{}-{}-{}'.format(self._server,
                                   self._handler,
                                   str(descriptor))
        hash_value = int.from_bytes(string.encode('utf-8'), 'little')
        return hash_value

    def find_duplicates(self, data_tasks: List[str],
                        descriptor: dict) -> List[str]:
        """Search for duplicates of the given descriptor and return a list
        of associated task ids.
        :param data_tasks: Limit duplicate search to.
        :param descriptor: ETL descriptor. Used to identify duplicates.
        :return: The list of duplicates.
        """
        task_ids = []
        hash_value = self.descriptor_to_hash(descriptor)
        for task_id in data_tasks:
            value = redis.get('data:{}'.format(task_id))
            if value is None:
                continue
            data_state = json.loads(value)
            if hash_value == data_state['hash']:
                task_ids.append(task_id)
        return task_ids

    def remove_duplicates(self, data_tasks: List[str],
                          descriptor: dict) -> None:
        """Delete the duplicates of the given descriptor from redis and call
        the janitor afterwards to cleanup orphaned files.
        :param data_tasks: Limit duplicate search to.
        :param descriptor: ETL descriptor. Used to identify duplicates.
        """
        task_ids = self.find_duplicates(data_tasks, descriptor)
        for task_id in task_ids:
            redis.delete('data:{}'.format(task_id))
        if task_ids:
            janitor.delay()

    def find_duplicate_task_id(self, data_tasks: List[str],
                               descriptor: dict) -> Union[str, None]:
        """Search for duplicates of the given descriptor and return their
        task id if the state is SUBMITTED or SUCCESS, meaning the data are
        reusable.
        :param data_tasks: Limit search to this list.
        :param descriptor: ETL descriptor. Used to identify duplicates.
        :return: TaskID if valid duplicate has been found, None otherwise.
        """
        task_ids = self.find_duplicates(data_tasks, descriptor)
        for task_id in task_ids:
            async_result = celery.AsyncResult(task_id)
            if (async_result.state == 'SUBMITTED' or
                    async_result.state == 'SUCCESS'):
                return task_id
        return None

    def handle(self, descriptors: List[dict], data_tasks: List[str],
               use_existing: bool, wait: bool = False) -> List[str]:
        """Create instances of ETL for the given descriptors and submit them
        (ETL implements celery.Task) to the broker. The task ids are returned
        to keep track of them.
        :param descriptors: A list of items describing the data to download.
        :param data_tasks: Limit search for duplicates to this list.
        :param use_existing: If a duplicate with state 'SUBMITTED' or 'SUCCESS'
        already exists use it instead of starting a new ETL. If this is False
        duplicates are deleted!
        :param wait: Makes this method synchronous by waiting for the tasks to
        return.
        :return: The list of task ids for the submitted tasks.
        """
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        task_ids = []
        for descriptor in descriptors:
            if use_existing:
                task_id = self.find_duplicate_task_id(data_tasks, descriptor)
                if task_id:
                    task_ids.append(task_id)
                    data_tasks.append(task_id)
                    continue
            else:
                self.remove_duplicates(data_tasks, descriptor)
            task_id = str(uuid4())
            file_path = os.path.join(data_dir, task_id)
            etl = ETL.factory(handler=self._handler, descriptor=descriptor)
            self.create_redis_entry(task_id, file_path,
                                    descriptor, etl.produces)
            kwargs = dict(server=self._server, token=self._token,
                          descriptor=descriptor, file_path=file_path,
                          encrypt=app.config['FRACTALIS_ENCRYPT_CACHE'])
            async_result = etl.apply_async(kwargs=kwargs, task_id=task_id)
            assert async_result.id == task_id
            task_ids.append(task_id)
            data_tasks.append(task_id)
            if wait and async_result.state == 'SUBMITTED':
                logger.debug("'wait' was set. Waiting for tasks to finish ...")
                async_result.get(propagate=False)
        task_ids = list(set(task_ids))
        return task_ids

    @staticmethod
    def factory(handler: str, server: str, auth: dict) -> 'ETLHandler':
        """Return an instance of the implementation of ETLHandler that can
        handle the given parameters.
        :param handler: Describes the handler. E.g.: transmart, ada
        :param server: The server to download data from.
        :param auth: Contains credentials to authenticate with the API.
        :return: An instance of an implementation of ETLHandler that returns
        True for self.can_handle
        """
        from . import HANDLER_REGISTRY
        for Handler in HANDLER_REGISTRY:
            if Handler.can_handle(handler):
                return Handler(server, auth)
        raise NotImplementedError(
            "No ETLHandler implementation found for: '{}'".format(handler))

    @classmethod
    def can_handle(cls, handler: str) -> bool:
        """Check whether this implementation of ETLHandler can handle the given
        parameters.
        :param handler: Describes the handler. E.g.: transmart, ada
        :return: True if this implementation can handle the given parameters.
        """
        return handler == cls._handler

    def _heartbeat(self):
        raise NotImplementedError()



"""This module provides the integrity checks
for the 'categorical' fractalis format."""

import logging

import numpy as np
import pandas as pd

from fractalis.data.check import IntegrityCheck

logger = logging.getLogger(__name__)


class CategoricalIntegrityCheck(IntegrityCheck):
    """Implements IntegrityCheck for 'categorical' data type."""

    data_type = 'categorical'

    def check(self, data: object) -> None:
        if not isinstance(data, pd.DataFrame):
            error = "Data must be a pandas.DataFrame."
            logger.error(error)
            raise ValueError(error)
        if sorted(['id', 'feature', 'value']) != sorted(data.columns.tolist()):
            error = "Data frame must contain the columns " \
                    "'id', 'feature', and 'value'."
            logger.error(error)
            raise ValueError(error)
        if data['id'].dtype != np.object:
            error = "'id' column must be of type 'object' ('string')."
            logger.error(error)
            raise ValueError(error)
        if data['feature'].dtype != np.object:
            error = "'feature' column must be of type 'object' ('string')."
            logger.error(error)
            raise ValueError(error)
        if data['value'].dtype != np.object:
            error = "'value' column must be of type 'object' ('string')."
            logger.error(error)
            raise ValueError(error)
        if len(data['id'].unique().tolist()) != data.shape[0]:
            error = "'id' column must be unique for this data type."
            logger.error(error)
            raise ValueError(error)
        if len(data['feature'].unique().tolist()) != 1:
            error = "'feature' column must contain exactly one unique value " \
                    "for this data type."
            logger.error(error)
            raise ValueError(error)



"""This module provides tests for the session interface of Fractalis."""

from uuid import UUID
from time import sleep

import pytest
import flask
from redis import StrictRedis


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestSession:

    @pytest.fixture(scope='module')
    def app(self):
        from fractalis import app
        app.testing = True
        return app

    @pytest.fixture(scope='module')
    def redis(self, app):
        redis = StrictRedis(host=app.config['REDIS_HOST'],
                            port=app.config['REDIS_PORT'])
        return redis

    def test_add_data_to_session_and_expect_it_in_db(self, app, redis):
        with app.test_client() as c:
            with c.session_transaction() as sess:
                sess.permanent = True
                sess['foo'] = 'bar'
                session_id = sess.sid
        assert redis.exists('session:{}'.format(session_id))

    def test_add_data_to_session_and_expect_sid_to_be_uuid(self, app):
        with app.test_client() as c:
            with c.session_transaction() as sess:
                sess.permanent = True
                sess['foo'] = 'bar'
                assert sess.sid
                UUID(sess.sid)

    def test_add_data_and_expect_cookie_set(self, app):
        with app.test_client() as c:
            with c.session_transaction() as sess:
                sess.permanent = True
                sess['foo'] = 'bar'
            c.get()
            assert flask.request.cookies

    def test_dont_add_data_and_exoect_no_cookie_set(self, app):
        with app.test_client() as c:
            c.get()
            assert not flask.request.cookies

    def test_change_session_data_and_expect_change_in_db(self, app, redis):
        with app.test_client() as c:
            with c.session_transaction() as sess:
                sess.permanent = True
                sess['foo'] = 'bar'
            session_id = sess.sid
        value_1 = redis.get('session:{}'.format(session_id))
        with app.test_client() as c:
            with c.session_transaction() as sess:
                sess.permanent = True
                sess['foo'] = 'baz'
            session_id = sess.sid
        value_2 = redis.get('session:{}'.format(session_id))
        assert value_1 != value_2

    def test_session_data_not_in_db_when_expired(self, app, redis):
        app.config['PERMANENT_SESSION_LIFETIME'] = 1
        with app.test_client() as c:
            with c.session_transaction() as sess:
                sess.permanent = True
                sess['foo'] = 'bar'
                session_id = sess.sid
        assert redis.exists('session:{}'.format(session_id))
        sleep(2)
        assert not redis.exists('session:{}'.format(session_id))



"""This module provides test for the highdim data ETL for tranSMART"""

import json

import pytest
import responses

from fractalis.data.etls.transmart.etl_highdim import HighdimETL


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
@pytest.mark.skip
class TestHighdimlETL:

    etl = HighdimETL()

    def test_correct_handler(self):
        assert self.etl.can_handle(handler='transmart',
                                   descriptor={'data_type': 'numerical_array'})
        assert not self.etl.can_handle(handler='ada',
                                       descriptor={'data_type': 'numerical_array'})
        assert not self.etl.can_handle(handler='ada',
                                       descriptor={'data_type': 'categorical'})
        assert not self.etl.can_handle(handler='ada',
                                       descriptor={'foo': 'bar'})

    def test_extract_raises_readable_if_not_200(self):
        with responses.RequestsMock() as response:
            response.add(response.GET, 'http://foo.bar/v2/observations',
                         body='{}',
                         status=400,
                         content_type='application/json')
            with pytest.raises(ValueError) as e:
                self.etl.extract(server='http://foo.bar',
                                 token='', descriptor={'path': ''})
                assert '[400]' in e

    def test_extract_raises_readable_if_not_json(self):
        with responses.RequestsMock() as response:
            response.add(response.GET, 'http://foo.bar/v2/observations',
                         body='123{//}',
                         status=200,
                         content_type='application/json')
            with pytest.raises(ValueError) as e:
                self.etl.extract(server='http://foo.bar',
                                 token='', descriptor={'path': ''})
                assert 'unexpected data' in e

    def test_extract_works_for_valid_input(self):
        with responses.RequestsMock() as response:
            response.add(response.GET, 'http://foo.bar/v2/observations',
                         body='{}',
                         status=200,
                         content_type='application/json')
            raw_data = self.etl.extract(server='http://foo.bar',
                                        token='', descriptor={'path': ''})
            assert isinstance(raw_data, dict)

    def test_transform_valid_input_correct_output(self):
        body = {
            "cells": [{"inlineDimensions": ["292278994-08-16T23:00:00Z", None, "@"], "dimensionIndexes": [0, 0, 0, None, 0, None, None], "numericValue": 52.0}],  # noqa: 501
            "dimensionElements": {"patient": [{"id": 1000421548, "deathDate": None, "birthDate": None, "race": None, "maritalStatus": None, "inTrialId": "3052", "age": 52, "trial": "GSE4382", "sexCd": None, "sex": "unknown", "religion": None}]}  # noqa: 501
        }
        with responses.RequestsMock() as response:
            response.add(response.GET, 'http://foo.bar/v2/observations',
                         body=json.dumps(body),
                         status=200,
                         content_type='application/json')
            raw_data = self.etl.extract(server='http://foo.bar',
                                        token='', descriptor={'path': ''})
            df = self.etl.transform(raw_data=raw_data, descriptor={'path': ''})
            assert df.shape == (1, 2)
            assert df.values.tolist() == [['3052', 52.0]]
            assert list(df) == ['id', 'value']



"""This module contains the tests for the Boxplot analysis code."""

import json

import numpy as np
import pandas as pd

from fractalis.analytics.tasks.boxplot.main import BoxplotTask


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestBoxplotAnalytics:

    task = BoxplotTask()

    def test_correct_output(self):
        df_1 = pd.DataFrame([[101, 'foo', 1], [102, 'foo', 2], [103, 'foo', 3],
                             [104, 'foo', 4]],
                            columns=['id', 'feature', 'value'])
        df_2 = pd.DataFrame([[101, 'bar', 1], [102, 'bar', 2], [103, 'bar', 3],
                             [104, 'bar', 4]],
                            columns=['id', 'feature', 'value'])
        results = self.task.main(features=[df_1, df_2],
                                 transformation='identity',
                                 categories=[],
                                 id_filter=[],
                                 subsets=[])
        json.dumps(results)  # check if result is json serializable
        assert 'data' in results
        assert 'statistics' in results
        assert 'anova' in results
        assert results['anova']['p_value'] == 1
        assert results['anova']['f_value'] == 0
        assert len(json.loads(results['data'])) == 8
        assert len(results['statistics']) == 2
        assert 'foo////s1' in results['statistics']
        assert 'bar////s1' in results['statistics']
        stats = results['statistics']['foo////s1']
        assert not np.isnan(stats['median'])
        assert not np.isnan(stats['l_qrt'])
        assert not np.isnan(stats['u_qrt'])
        assert not np.isnan(stats['l_wsk'])
        assert not np.isnan(stats['u_wsk'])

    def test_marks_outliers(self):
        df_1 = pd.DataFrame([[100, 'foo', -50],
                             [101, 'foo', 1],
                             [102, 'foo', 2],
                             [103, 'foo', 3],
                             [104, 'foo', 100]],
                            columns=['id', 'feature', 'value'])
        df_2 = pd.DataFrame([[201, 'bar', 1],
                             [202, 'bar', 2],
                             [203, 'bar', 3],
                             [204, 'bar', 100]],
                            columns=['id', 'feature', 'value'])
        results = self.task.main(features=[df_1, df_2], categories=[],
                                 transformation='identity',
                                 id_filter=[], subsets=[])
        df = pd.DataFrame.from_dict(json.loads(results['data']))
        assert np.all(df['outlier'] == [True, False, False, False, True,
                                        False, False, False, True])

    def test_can_handle_nan(self):
        df = pd.DataFrame([[100, 'foo', -50],
                           [101, 'foo', 1],
                           [102, 'foo', float('nan')],
                           [103, 'foo', 3],
                           [104, 'foo', 100]],
                          columns=['id', 'feature', 'value'])
        results = self.task.main(features=[df], categories=[],
                                 transformation='identity',
                                 id_filter=[], subsets=[])
        assert results['statistics']['foo////s1']['median'] == 2

    def test_can_handle_groups_with_only_nan(self):
        df = pd.DataFrame([[100, 'foo', -50],
                           [101, 'foo', 1],
                           [102, 'foo', float('nan')],
                           [103, 'foo', 3],
                           [104, 'foo', 100],
                           [105, 'foo', float('nan')]],
                          columns=['id', 'feature', 'value'])
        categories = pd.DataFrame([[100, 'gender', 'female'],
                                   [101, 'gender', 'female'],
                                   [102, 'gender', 'male'],
                                   [103, 'gender', 'female'],
                                   [104, 'gender', 'female'],
                                   [105, 'gender', 'male']],
                                  columns=['id', 'feature', 'value'])
        results = self.task.main(features=[df], categories=[categories],
                                 transformation='identity',
                                 id_filter=[], subsets=[])
        assert 'foo//female//s1' in results['statistics']
        assert 'foo//male//s1' not in results['statistics']



"""This module provides PicSureHandler,
an implementation of ETLHandler for PIC-SURE."""

import logging

from fractalis.data.etlhandler import ETLHandler


logger = logging.getLogger(__name__)


class PicSureHandler(ETLHandler):
    """This ETLHandler provides integration with PIC-SURE."""

    _handler = 'pic-sure'

    @staticmethod
    def make_label(descriptor: dict) -> str:
        return descriptor['query']['select'][0]['alias']

    def _get_token_for_credentials(self, server: str, auth: dict) -> str:
        return auth['token']



"""This module provides tests for the integer ETL for Ada."""

import json

import pytest
import responses

from fractalis.data.etls.ada.etl_integer import IntegerETL


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestIntegerETL:

    etl = IntegerETL()

    valid_descriptor = {
        'dictionary': {
            'name': 'foo',
            'projection': 'foo',
            'label': 'bar',
            'fieldType': 'Integer',
            'isArray': False
        },
        'data_set': 'baz'
    }

    def test_correct_handler(self):
        assert self.etl.can_handle(handler='ada', descriptor={
            'dictionary': {'fieldType': 'Integer', 'isArray': False}
        })
        assert not self.etl.can_handle(handler='ada', descriptor={
            'dictionary': {'fieldType': 'Integer', 'isArray': True}
        })
        assert not self.etl.can_handle(handler='transmart', descriptor={
            'dictionary': {'fieldType': 'Integer', 'isArray': False}
        })
        assert not self.etl.can_handle(handler='ada', descriptor={
            'dictionary': {'fieldType': 'foo', 'isArray': False}
        })

    def test_extract_raises_readable_if_not_200(self):
        with responses.RequestsMock() as response:
            response.add(response.GET,
                         'http://foo.bar/dataSets/records/findCustom',
                         body='{}',
                         status=400,
                         content_type='application/json')
            with pytest.raises(ValueError) as e:
                self.etl.extract(server='http://foo.bar',
                                 token='',
                                 descriptor=self.valid_descriptor)
                assert '[400]' in e

    def test_extract_raises_readable_if_not_json(self):
        with responses.RequestsMock() as response:
            response.add(response.GET,
                         'http://foo.bar/dataSets/records/findCustom',
                         body='abc{//}',
                         status=200,
                         content_type='application/json')
            with pytest.raises(TypeError) as e:
                self.etl.extract(server='http://foo.bar',
                                 token='',
                                 descriptor=self.valid_descriptor)
                assert 'unexpected data' in e

    def test_returns_data_for_correct_input(self):
        body = json.dumps([{'a': 1, '_id': {'$oid': '12345'}}])
        with responses.RequestsMock() as response:
            response.add(response.GET,
                         'http://foo.bar/dataSets/records/findCustom',
                         body=body,
                         status=200,
                         content_type='application/json')
            raw_data = self.etl.extract(server='http://foo.bar',
                                        token='',
                                        descriptor=self.valid_descriptor)
            assert isinstance(raw_data, list)

    def test_returns_df_for_correct_data(self):
        body = json.dumps([{'a': 1, '_id': {'$oid': '12345'}}])
        with responses.RequestsMock() as response:
            response.add(response.GET,
                         'http://foo.bar/dataSets/records/findCustom',
                         body=body,
                         status=200,
                         content_type='application/json')
            raw_data = self.etl.extract(server='http://foo.bar',
                                        token='',
                                        descriptor=self.valid_descriptor)
            df = self.etl.transform(raw_data=raw_data,
                                    descriptor=self.valid_descriptor)
            assert df.shape == (1, 3)
            assert list(df['id']) == ['12345']
            assert list(df['feature']) == ['a']
            assert list(df['value']) == [1]



"""The /analytics controller. Please refer to doc/api for more information."""

import logging
from typing import Tuple
from uuid import UUID

from flask import Blueprint, session, request, jsonify
from flask.wrappers import Response

from fractalis import celery, app
from fractalis.validator import validate_json, validate_schema
from fractalis.analytics.schema import create_task_schema
from fractalis.analytics.task import AnalyticTask


analytics_blueprint = Blueprint('analytics_blueprint', __name__)
logger = logging.getLogger(__name__)


@analytics_blueprint.route('', methods=['POST'])
@validate_json
@validate_schema(create_task_schema)
def create_task() -> Tuple[Response, int]:
    """Create a new analytics task based on the parameters in the POST body.
    See doc/api/ for more information.
    :return: Flask Response
    """
    logger.debug("Received POST request on /analytics.")
    json = request.get_json(force=True)
    analytic_task = AnalyticTask.factory(json['task_name'])
    if analytic_task is None:
        logger.error("Could not submit task for unknown task name: "
                     "'{}'".format(json['task_name']))
        return jsonify({'error_msg': "Task with name '{}' not found."
                       .format(json['task_name'])}), 400
    async_result = analytic_task.delay(
        session_data_tasks=session['data_tasks'], args=json['args'],
        decrypt=app.config['FRACTALIS_ENCRYPT_CACHE'])
    session['analytic_tasks'].append(async_result.id)
    logger.debug("Task successfully submitted. Sending response.")
    return jsonify({'task_id': async_result.id}), 201


@analytics_blueprint.route('/<uuid:task_id>', methods=['GET'])
def get_task_details(task_id: UUID) -> Tuple[Response, int]:
    """Get task details for the given task_id.
     See doc/api/ for more information.
    :param task_id: ID returned on task creation.
    :return: Flask Response
    """
    logger.debug("Received GET request on /analytics/task_id.")
    wait = request.args.get('wait') == '1'
    task_id = str(task_id)
    if task_id not in session['analytic_tasks']:
        error = "Task ID '{}' not found in session. " \
                "Refusing access.".format(task_id)
        logger.warning(error)
        return jsonify({'error': error}), 403
    async_result = celery.AsyncResult(task_id)
    if wait and async_result.state == 'SUBMITTED':
        async_result.get(propagate=False)
    result = async_result.result
    if isinstance(result, Exception):  # Exception -> str
        result = "{}: {}".format(type(result).__name__, str(result))
    logger.debug("Task found and has access. Sending response.")
    return jsonify({'state': async_result.state, 'result': result}), 200


@analytics_blueprint.route('/<uuid:task_id>', methods=['DELETE'])
def cancel_task(task_id: UUID) -> Tuple[Response, int]:
    """Cancel a task for a given task_id.
    See doc/api/ for more information.
    :param task_id: ID returned on task creation.
    :return: Flask Response
    """
    logger.debug("Received DELETE request on /analytics/task_id.")
    task_id = str(task_id)
    if task_id not in session['analytic_tasks']:
        error = "Task ID '{}' not found in session. " \
                "Refusing access.".format(task_id)
        logger.warning(error)
        return jsonify({'error': error}), 403
    wait = request.args.get('wait') == '1'
    # possibly dangerous: http://stackoverflow.com/a/29627549
    celery.control.revoke(task_id, terminate=True, signal='SIGUSR1', wait=wait)
    session['analytic_tasks'].remove(task_id)
    logger.debug("Successfully send term signal to task. Sending response.")
    return jsonify(''), 200



import json
import logging
from uuid import uuid4
from time import sleep

from werkzeug.datastructures import CallbackDict
from flask import session
from flask.sessions import SessionMixin, SessionInterface


logger = logging.getLogger(__name__)


class RedisSession(CallbackDict, SessionMixin):

    def __init__(self, sid, initial=None):
        if initial is None:
            initial = {
                'data_tasks': [],
                'analytic_tasks': [],
                'subsets': [],
                'state_access': {}
            }

        def on_update(self):
            self.modified = True
        CallbackDict.__init__(self, initial, on_update)
        self.sid = sid
        self.permanent = True
        self.modified = False


class RedisSessionInterface(SessionInterface):

    def __init__(self, redis, app):
        self.redis = redis

        @app.teardown_request
        def teardown_request(exc=None) -> None:
            """Release session lock whatever happens.
            :param exc: Unhandled exception that might have been thrown.
            """
            try:
                app.session_interface.release_lock(session.sid)
            except AttributeError:
                logger.warning(
                    "Attempted to release session lock but no session id was "
                    "found. This happens only during testing and should never "
                    "happen in production mode.")

    def acquire_lock(self, sid, request_id):
        if request_id is None:
            return
        lock_id = self.redis.get(name='session:{}:lock'.format(sid))
        if lock_id == request_id:
            return
        while self.redis.getset(name='session:{}:lock'.format(sid),
                                value=request_id):
            sleep(0.1)
        self.redis.setex(name='session:{}:lock'.format(sid),
                         value=request_id, time=10)

    def release_lock(self, sid):
        self.redis.delete('session:{}:lock'.format(sid))

    def open_session(self, app, request):
        request_id = request.environ.get("FLASK_REQUEST_ID")
        sid = request.cookies.get(app.session_cookie_name)
        if not sid:
            sid = str(uuid4())
            self.acquire_lock(sid, request_id)
            return RedisSession(sid=sid)
        self.acquire_lock(sid, request_id)
        session_data = self.redis.get('session:{}'.format(sid))
        if session_data is not None:
            session_data = json.loads(session_data)
            return RedisSession(sid=sid, initial=session_data)
        return RedisSession(sid=sid)

    def save_session(self, app, session, response):
        path = self.get_cookie_path(app)
        domain = self.get_cookie_domain(app)
        if not session:
            if session.modified:
                self.redis.delete('session:{}'.format(session.sid))
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain, path=path)
            return
        session_expiration_time = app.config['PERMANENT_SESSION_LIFETIME']
        cookie_expiration_time = self.get_expiration_time(app, session)
        serialzed_session_data = json.dumps(dict(session))
        self.redis.setex(name='session:{}'.format(session.sid),
                         time=session_expiration_time,
                         value=serialzed_session_data)
        response.set_cookie(key=app.session_cookie_name, value=session.sid,
                            expires=cookie_expiration_time, httponly=True,
                            domain=domain)



"""Initialize Fractalis Flask app and configure it."""

import logging.config
import yaml
from flask import Flask
from flask_cors import CORS
from flask_request_id import RequestID
from flask_compress import Compress
from redis import StrictRedis

from fractalis.session import RedisSessionInterface

app = Flask(__name__)

# Configure app with defaults
app.config.from_object('fractalis.config')
# Configure app with manually settings
default_config = True
try:
    app.config.from_envvar('FRACTALIS_CONFIG')
    default_config = False
except RuntimeError:
    pass
app.config['PERMANENT_SESSION_LIFETIME'] =\
    app.config['FRACTALIS_DATA_LIFETIME']
app.config['CELERY_TASK_RESULT_EXPIRES'] =\
    app.config['FRACTALIS_DATA_LIFETIME']

# setup logging
with open(app.config['FRACTALIS_LOG_CONFIG'], 'rt') as f:
    log_config = yaml.safe_load(f.read())
logging.config.dictConfig(log_config)
log_level = logging._nameToLevel[log_config['handlers']['default']['level']]
logging.getLogger('werkzeug').setLevel(log_level)
log = logging.getLogger(__name__)

# we can't log this earlier because the logger depends on the loaded app config
if default_config:
    log.warning("Environment Variable FRACTALIS_CONFIG not set. Falling back "
                "to default settings. This is not a good idea in production!")

# Plugin that assigns every request an id
RequestID(app)

# Plugin that compresses all responses
Compress(app)

# create a redis instance
log.info("Creating Redis connection.")
redis = StrictRedis(host=app.config['REDIS_HOST'],
                    port=app.config['REDIS_PORT'],
                    charset='utf-8',
                    decode_responses=True)

# Set new session interface for app
log.info("Replacing default session interface.")
app.session_interface = RedisSessionInterface(redis, app)

# allow everyone to submit requests
log.info("Setting up CORS.")
CORS(app, supports_credentials=True)

# create celery app
log.info("Creating celery app.")
from fractalis.celeryapp import make_celery, register_tasks  # noqa: E402
celery = make_celery(app)

# register blueprints
from fractalis.analytics.controller import analytics_blueprint  # noqa: E402
from fractalis.data.controller import data_blueprint  # noqa: E402
from fractalis.misc.controller import misc_blueprint  # noqa: E402
from fractalis.state.controller import state_blueprint  # noqa: E402
log.info("Registering Flask blueprints.")
app.register_blueprint(analytics_blueprint, url_prefix='/analytics')
app.register_blueprint(data_blueprint, url_prefix='/data')
app.register_blueprint(misc_blueprint, url_prefix='/misc')
app.register_blueprint(state_blueprint, url_prefix='/state')

# registering all application celery tasks
log.info("Registering celery tasks.")
register_tasks()

log.info("Initialisation of service complete.")

if __name__ == '__main__':
    log.info("Starting builtin web server.")
    app.run(host='0.0.0.0', port=5000)
    log.info("Builtin web server started.")



"""Provides DoubleArrayETL for Ada."""

from typing import List

import pandas as pd

from fractalis.data.etl import ETL
from fractalis.data.etls.ada import shared


class DoubleArrayETL(ETL):
    """DoubleArrayETL implements support for Adas 'Enum' array type."""

    name = 'ada_double_array_etl'
    produces = 'numerical_array'

    @staticmethod
    def can_handle(handler, descriptor):
        return handler == 'ada' and \
               descriptor['dictionary']['fieldType'] and \
               descriptor['dictionary']['fieldType'] == 'Double' and \
               descriptor['dictionary']['isArray']

    def extract(self, server: str, token: str, descriptor: dict) -> List[dict]:
        data_set = descriptor['data_set']
        projection = descriptor['dictionary']['projection']
        cookie = shared.make_cookie(token=token)
        data = shared.get_field(server=server, data_set=data_set,
                                cookie=cookie, projection=projection)
        return data

    def transform(self, raw_data: List[dict],
                  descriptor: dict) -> pd.DataFrame:
        name = descriptor['dictionary']['name']
        ids = []
        values = []
        for row in raw_data:
            ids.append(row['_id']['$oid'])
            values.append(row[name])
        df = pd.DataFrame(values)
        df.insert(0, 'id', ids)
        df = pd.melt(df, id_vars='id', var_name='feature')
        return df



import os
from datetime import timedelta

# DO NOT MODIFY THIS FILE DIRECTLY


# Flask
SECRET_KEY = 'OVERWRITE ME IN PRODUCTION!!!'
DEBUG = False
TESTING = False
REDIS_HOST = 'localhost'
REDIS_PORT = '6379'
SESSION_COOKIE_HTTPONLY = True
SESSION_COOKIE_SECURE = False
SESSION_REFRESH_EACH_REQUEST = True

# Celery
BROKER_URL = 'amqp://guest:guest@localhost:5672//'
CELERY_RESULT_BACKEND = 'redis://localhost:6379'
CELERYD_TASK_SOFT_TIME_LIMIT = 60 * 20
CELERYD_TASK_TIME_LIMIT = 60 * 30
CELERYD_HIJACK_ROOT_LOGGER = False

# Fractalis
# Location of cache and temporary files
FRACTALIS_TMP_DIR = os.path.abspath(os.path.join(os.sep, 'tmp', 'fractalis'))
# How long to store files in the cache
FRACTALIS_DATA_LIFETIME = timedelta(days=6)
# How long to keep analysis results (beware of high RAM usage)
FRACTALIS_RESULT_LIFETIME = timedelta(seconds=30)
# Should the Cache be encrypted? This might impact performance for little gain!
FRACTALIS_ENCRYPT_CACHE = False
# Location of your the log configuration file.
FRACTALIS_LOG_CONFIG = os.path.join(os.path.dirname(__file__), 'logging.yaml')
# Whether to verify the certs of https data sources
ETL_VERIFY_SSL_CERT = False

# DO NOT MODIFY THIS FILE DIRECTLY



"""The /misc controller provides an API for everything that does not belong
in any of the other categories."""

import logging
from typing import Tuple

from flask import Blueprint, jsonify, Response

from fractalis.cleanup import janitor


misc_blueprint = Blueprint('misc_blueprint', __name__)
logger = logging.getLogger(__name__)


@misc_blueprint.route('/version', methods=['GET'])
def get_version() -> Tuple[Response, int]:
    version = '1.1.0'
    # this is a good place to launch the janitor because /version is one of the
    # first requests sent by the front-end on initialization
    janitor.delay()
    return jsonify({'version': version}), 201



"""This module provides tests for the integrity checker
for the 'categorical' data type."""

import pandas as pd
import pytest

from fractalis.data.check import IntegrityCheck


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestCategoricalIntegrityCheck:

    checker = IntegrityCheck.factory('categorical')

    def test_correct_check_1(self):
        df = pd.DataFrame([['1', '2', '3']],
                          columns=['id', 'feature', 'value'])
        self.checker.check(df)

    def test_correct_check_2(self):
        df = pd.DataFrame([['1', '2', '3']], columns=['id', 'feat', 'value'])
        with pytest.raises(ValueError) as e:
            self.checker.check(df)
            assert 'must contain the columns' in e

    def test_correct_check_3(self):
        df = pd.DataFrame([[1, '2', '3']], columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            self.checker.check(df)
            assert "must be of type 'object'" in e

    def test_correct_check_4(self):
        df = pd.DataFrame([['1', 2, '3']], columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            self.checker.check(df)
            assert "must be of type 'object'" in e

    def test_correct_check_5(self):
        df = pd.DataFrame([['1', '2', 3]], columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            self.checker.check(df)
            assert "must be of type 'object'" in e

    def test_correct_check_6(self):
        df = pd.DataFrame([['1', '2', '3'], ['4', '2', '3']],
                          columns=['id', 'feature', 'value'])
        self.checker.check(df)

    def test_correct_check_7(self):
        df = pd.DataFrame([['1', '2', '3'], ['4', '4', '3']],
                          columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            self.checker.check(df)
            assert 'must contain exactly one' in e

    def test_correct_check_8(self):
        df = pd.DataFrame([['1', '2', '3'], ['1', '2', '3']],
                          columns=['id', 'feature', 'value'])
        with pytest.raises(ValueError) as e:
            self.checker.check(df)
            assert 'must be unique' in e



"""This module tests the state controller module."""

import json

from uuid import UUID, uuid4
import flask
import pytest

from fractalis import redis, sync, celery
from fractalis.data.etlhandler import ETLHandler


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestState:

    @pytest.fixture(scope='function')
    def test_client(self):
        sync.cleanup_all()
        from fractalis import app
        app.testing = True
        with app.test_client() as test_client:
            yield test_client
            sync.cleanup_all()

    def test_400_if_no_task_id_in_payload(self, test_client):
        payload = {
            'state': {'abc': '$...foo'},
            'handler': 'test',
            'server': 'localfoo'
        }
        rv = test_client.post('/state', data=flask.json.dumps(payload))
        body = flask.json.loads(rv.get_data())
        assert 400 == rv.status_code, body
        assert 'error' in body
        assert 'contains no data task ids' in body['error']

    def test_400_if_task_id_not_in_redis(self, test_client):
        payload = {
            'state': {'abc': '${}$'.format(uuid4())},
            'handler': 'test',
            'server': 'localfoo'
        }
        rv = test_client.post('/state', data=flask.json.dumps(payload))
        body = flask.json.loads(rv.get_data())
        assert 400 == rv.status_code, body
        assert 'error' in body
        assert 'could not be found in redis' in body['error']

    def test_save_state_saves_and_returns(self, test_client):
        uuid = str(uuid4())
        payload = {
            'state': {'test': ['${}$'.format(uuid), '${${}']},
            'handler': 'test',
            'server': 'localfoo'
        }
        redis.set(name='data:{}'.format(uuid),
                  value=json.dumps({'meta': {'descriptor': 'foo'}}))
        rv = test_client.post('/state', data=flask.json.dumps(payload))
        body = flask.json.loads(rv.get_data())
        assert 201 == rv.status_code, body
        assert UUID(body['state_id'])
        meta_state = json.loads(redis.get('state:{}'.format(body['state_id'])))
        assert meta_state['task_ids'] == [uuid]
        assert meta_state['state']['test'][0] == '${}$'.format(uuid)

    def test_save_state_discards_duplicates(self, test_client):
        uuid1 = str(uuid4())
        uuid2 = str(uuid4())
        payload = {
            'state': {'test': ['${}$'.format(uuid1),
                               '${}$'.format(uuid1),
                               '${}$'.format(uuid2)]},
            'handler': 'test',
            'server': 'localfoo'
        }
        redis.set(name='data:{}'.format(uuid1),
                  value=json.dumps({'meta': {'descriptor': 'foo'}}))
        redis.set(name='data:{}'.format(uuid2),
                  value=json.dumps({'meta': {'descriptor': 'bar'}}))
        rv = test_client.post('/state', data=flask.json.dumps(payload))
        body = flask.json.loads(rv.get_data())
        assert 201 == rv.status_code, body
        assert UUID(body['state_id'])
        meta_state = json.loads(redis.get('state:{}'.format(body['state_id'])))
        assert len(meta_state['task_ids']) == 2
        assert len(meta_state['descriptors']) == 2
        assert uuid1 in meta_state['task_ids']
        assert uuid2 in meta_state['task_ids']
        assert 'foo' in meta_state['descriptors']
        assert 'bar' in meta_state['descriptors']

    def test_400_if_payload_schema_incorrect_1(self, test_client):
        payload = {
            'state': {'test': ['${}$'.format(uuid4())]},
            'server': 'localfoo'
        }
        rv = test_client.post('/state', data=flask.json.dumps(payload))
        assert 400 == rv.status_code

    def test_404_if_request_invalid_state_id(self, test_client):
        rv = test_client.post('/state/{}'.format(str(uuid4())),
                              data=flask.json.dumps({'auth': {'token': ''}}))
        assert 404 == rv.status_code
        body = flask.json.loads(rv.get_data())
        assert 'error' in body
        assert 'not find state associated with id' in body['error']

    def test_400_if_payload_schema_incorrect_2(self, test_client):
        payload = {
            'auth': {}
        }
        rv = test_client.post('/state/{}'.format(str(uuid4())),
                              data=flask.json.dumps(payload))
        assert 400 == rv.status_code

    def test_request_state_acces_works(self, test_client):
        meta_state = {
            'state': {'foo': ['$123$', '$456$']},
            'handler': 'test',
            'server': 'localfoo',
            'task_ids': ['123', '456'],
            'descriptors': [
                {'data_type': 'default'},
                {'data_type': 'default', 'foo': 'bar'}
            ],
        }
        uuid = str(uuid4())
        redis.set(name='state:{}'.format(uuid),
                  value=json.dumps(meta_state))
        with test_client.session_transaction() as sess:
            assert not sess['data_tasks']
            assert not sess['state_access']
        rv = test_client.post('/state/{}'.format(uuid),
                              data=flask.json.dumps({'auth': {'token': ''}}))
        body = flask.json.loads(rv.get_data())
        assert 202 == rv.status_code, body
        assert not body
        with test_client.session_transaction() as sess:
            assert len(sess['data_tasks']) == 2
            assert len(sess['state_access']) == 1
            key = list(sess['state_access'].keys())[0]
            assert len(sess['state_access'][key]) == 2
            assert sess['data_tasks'][0] in sess['state_access'][key]
            assert sess['data_tasks'][1] in sess['state_access'][key]
            assert meta_state['task_ids'][0] not in sess['state_access'][key]
            assert meta_state['task_ids'][1] not in sess['state_access'][key]

    def test_request_state_access_reuses_duplicate(
            self, test_client):
        uuid1 = str(uuid4())
        uuid2 = str(uuid4())
        meta_state = {
            'state': {'foo': ['${}$'.format(uuid1), '${}$'.format(uuid2)]},
            'handler': 'test',
            'server': 'localfoo',
            'task_ids': [uuid1, uuid2],
            'descriptors': [
                {'data_type': 'default'},
                {'data_type': 'default'}
            ],
        }
        uuid = str(uuid4())
        redis.set(name='state:{}'.format(uuid),
                  value=json.dumps(meta_state))
        with test_client.session_transaction() as sess:
            assert not sess['data_tasks']
            assert not sess['state_access']
        rv = test_client.post('/state/{}'.format(uuid),
                              data=flask.json.dumps({'auth': {'token': ''}}))
        body = flask.json.loads(rv.get_data())
        assert 202 == rv.status_code, body
        assert not body
        with test_client.session_transaction() as sess:
            assert len(sess['data_tasks']) == 1
            assert len(sess['state_access']) == 1
            key = list(sess['state_access'].keys())[0]
            assert len(sess['state_access'][key]) == 1
            assert sess['data_tasks'] == sess['state_access'][key]
            assert meta_state['task_ids'][0] != sess['state_access'][key][0]

    def test_request_state_reuses_previous_etls_but_only_in_own_scope(
            self, test_client, monkeypatch):
        uuid1 = str(uuid4())
        uuid2 = str(uuid4())
        descriptor_1 = {'data_type': 'default', 'id': 1}
        descriptor_2 = {'data_type': 'default', 'id': 2}
        handler = 'test'
        server = 'localfoo'
        meta_state = {
            'state': {'foo': ['${}$'.format(uuid1), '${}$'.format(uuid2)]},
            'handler': handler,
            'server': server,
            'task_ids': [uuid1, uuid2],
            'descriptors': [
                descriptor_1,
                descriptor_2
            ],
        }
        uuid = str(uuid4())
        redis.set(name='state:{}'.format(uuid),
                  value=json.dumps(meta_state))
        etlhandler = ETLHandler.factory(handler=handler,
                                        server=server,
                                        auth={})
        etlhandler.create_redis_entry(task_id=uuid1,
                                      file_path='',
                                      descriptor=descriptor_1,
                                      data_type='')
        etlhandler.create_redis_entry(task_id=uuid2,
                                      file_path='',
                                      descriptor=descriptor_2,
                                      data_type='')
        with test_client.session_transaction() as sess:
            sess['data_tasks'] = [uuid2]

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUCCESS'
                self.id = args[0]

            def get(self, *args, **kwargs):
                pass

        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        rv = test_client.post('/state/{}'.format(uuid),
                              data=flask.json.dumps({'auth': {'token': ''}}))
        body = flask.json.loads(rv.get_data())
        assert 202 == rv.status_code, body
        assert not body
        with test_client.session_transaction() as sess:
            assert len(sess['data_tasks']) == 2
            key = list(sess['state_access'].keys())[0]
            assert len(sess['state_access'][key]) == 2
            assert meta_state['task_ids'][0] not in sess['state_access'][key]
            assert meta_state['task_ids'][1] in sess['state_access'][key]

    def test_get_state_data_404_if_not_requested_before(self, test_client):
        rv = test_client.get('/state/{}'.format(str(uuid4())))
        assert 404 == rv.status_code
        body = flask.json.loads(rv.get_data())
        assert 'Cannot get state' in body.get('error')

    def test_get_state_with_replaced_ids_if_all_tasks_succeed(
            self, test_client, monkeypatch):
        meta_state = {
            'state': {'foo': ['$123$', '$456$']},
            'task_ids': ['123', '456'],
        }
        uuid = str(uuid4())
        redis.set(name='state:{}'.format(uuid),
                  value=json.dumps(meta_state))
        with test_client.session_transaction() as sess:
            sess['state_access'][uuid] = ['abc', 'efg']

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUCCESS'
                self.id = args[0]

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        rv = test_client.get('/state/{}'.format(uuid))
        body = flask.json.loads(rv.get_data())
        assert 200 == rv.status_code, body
        assert body['state']['foo'][0] == '$abc$'
        assert body['state']['foo'][1] == '$efg$'

    def test_get_state_get_message_if_not_all_tasks_finished(
            self, test_client, monkeypatch):
        meta_state = {
            'state': {'foo': ['$123$', '$456$']},
            'task_ids': ['123', '456'],
        }
        uuid = str(uuid4())
        redis.set(name='state:{}'.format(uuid),
                  value=json.dumps(meta_state))
        with test_client.session_transaction() as sess:
            sess['state_access'][uuid] = ['abc', 'efg']

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUBMITTED'
                self.id = args[0]

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        rv = test_client.get('/state/{}'.format(uuid))
        body = flask.json.loads(rv.get_data())
        assert 202 == rv.status_code, body
        assert 'still running' in body.get('message')

    def test_get_state_refuse_if_one_task_fails(
            self, test_client, monkeypatch):
        meta_state = {
            'state': {'foo': ['$123$', '$456$']},
            'task_ids': ['123', '456'],
        }
        uuid = str(uuid4())
        redis.set(name='state:{}'.format(uuid),
                  value=json.dumps(meta_state))
        with test_client.session_transaction() as sess:
            sess['state_access'][uuid] = ['abc', 'efg']

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'FAILURE'
                self.id = args[0]

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        rv = test_client.get('/state/{}'.format(uuid))
        body = flask.json.loads(rv.get_data())
        assert 403 == rv.status_code, body
        assert 'no access' in body.get('error')



"""This module provides statistics for a Survival Analysis."""

import logging
from typing import List

import pandas as pd
import numpy as np
from lifelines import KaplanMeierFitter, NelsonAalenFitter

from fractalis.analytics.task import AnalyticTask
from fractalis.analytics.tasks.shared import utils


logger = logging.getLogger(__name__)


class SurvivalTask(AnalyticTask):
    """Survival Analysis Task implementing AnalyticTask.
    This class is a submittable celery task."""

    name = 'survival-analysis'

    def main(self, durations: List[pd.DataFrame],
             categories: List[pd.DataFrame],
             event_observed: List[pd.DataFrame],
             estimator: str,
             id_filter: List[str],
             subsets: List[List[str]]) -> dict:
        # TODO: Docstring
        if len(durations) != 1:
            error = 'Analysis requires exactly one array that specifies the ' \
                    'duration length.'
            logger.exception(error)
            raise ValueError(error)
        if len(event_observed) > 1:
            error = 'Maximal one variable for "event_observed" allowed'
            logger.exception(error)
            raise ValueError(error)

        df = durations[0]
        df.dropna(inplace=True)
        if id_filter:
            df = df[df['id'].isin(id_filter)]
        df = utils.apply_subsets(df=df, subsets=subsets)
        df = utils.apply_categories(df=df, categories=categories)

        stats = {}
        categories = df['category'].unique().tolist()
        subsets = df['subset'].unique().tolist()
        # for every category and subset combination estimate the survival fun.
        for category in categories:
            for subset in subsets:
                sub_df = df[(df['category'] == category) &
                            (df['subset'] == subset)]
                T = sub_df['value']
                E = None  # default is nothing is censored
                if len(T) <= 3:
                    continue
                if event_observed:
                    # find observation boolean value for every duration
                    E = event_observed[0].merge(sub_df, how='right', on='id')
                    E = [bool(x) and not np.isnan(x) for x in E['value']]
                    assert len(E) == len(T)
                if estimator == 'NelsonAalen':
                    fitter = NelsonAalenFitter()
                    fitter.fit(durations=T, event_observed=E)
                    estimate = fitter.cumulative_hazard_[
                        'NA_estimate'].tolist()
                    ci_lower = fitter.confidence_interval_[
                        'NA_estimate_lower_0.95'].tolist()
                    ci_upper = fitter.confidence_interval_[
                        'NA_estimate_upper_0.95'].tolist()
                elif estimator == 'KaplanMeier':
                    fitter = KaplanMeierFitter()
                    fitter.fit(durations=T, event_observed=E)
                    # noinspection PyUnresolvedReferences
                    estimate = fitter.survival_function_[
                        'KM_estimate'].tolist()
                    ci_lower = fitter.confidence_interval_[
                        'KM_estimate_lower_0.95'].tolist()
                    ci_upper = fitter.confidence_interval_[
                        'KM_estimate_upper_0.95'].tolist()
                else:
                    error = 'Unknown estimator: {}'.format(estimator)
                    logger.exception(error)
                    raise ValueError(error)
                timeline = fitter.timeline.tolist()
                if not stats.get(category):
                    stats[category] = {}
                stats[category][subset] = {
                    'timeline': timeline,
                    'estimate': estimate,
                    'ci_lower': ci_lower,
                    'ci_upper': ci_upper
                }

        return {
            'label': df['feature'].tolist()[0],
            'categories': categories,
            'subsets': subsets,
            'stats': stats
        }



"""Module containing analysis code for pca."""

from typing import List, TypeVar
from functools import reduce
import logging

import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import Imputer

from fractalis.analytics.task import AnalyticTask
from fractalis.analytics.tasks.shared import utils


T = TypeVar('T')
logger = logging.getLogger(__name__)


class PCATask(AnalyticTask):
    """PCATask implementing AnalyticsTask. This class is a
    submittable celery task."""

    name = 'compute-pca'

    def main(self,
             features: List[pd.DataFrame],
             categories: List[pd.DataFrame],
             whiten: bool,
             id_filter: List[T],
             subsets: List[List[T]]) -> dict:
        # merge input data into single df
        df = reduce(lambda a, b: a.append(b), features)
        if not subsets:
            # empty subsets equals all samples in one subset
            subsets = [df['id'].unique().tolist()]

        # make matrix of data
        df = df.pivot(index='feature', columns='id', values='value')
        df = df.T
        feature_labels = list(df)

        # apply id filter
        if id_filter:
            df = df[df.index.isin(id_filter)]

        # save ids so we can re-assign them after pca
        ids = df.index.tolist()

        # replace missing values with row medians
        imp = Imputer(missing_values='NaN', strategy='median', axis=0)
        imp.fit(df)
        df = imp.transform(df)

        # PCA
        pca = PCA(whiten=whiten)
        pca.fit(df)
        reduced_df = pca.transform(df)

        # get explained variance ratios of components
        variance_ratios = pca.explained_variance_ratio_

        # get loadings
        loadings = -1 * pca.components_.T * np.sqrt(pca.explained_variance_)
        loadings = pd.DataFrame(loadings)
        loadings['feature'] = feature_labels

        # re-assign ids
        reduced_df = pd.DataFrame(reduced_df)
        reduced_df['id'] = ids

        # add category and subset column
        reduced_df = utils.apply_subsets(df=reduced_df, subsets=subsets)
        reduced_df = utils.apply_categories(df=reduced_df,
                                            categories=categories)

        return {
            'data': reduced_df.to_dict(orient='list'),
            'loadings': loadings.to_dict(orient='list'),
            'variance_ratios': variance_ratios.tolist()
        }



"""This module tests the data controller module."""

import json
import os
from uuid import UUID, uuid4

import flask
import pytest

from fractalis import app, redis, sync


# noinspection PyMissingOrEmptyDocstring, PyMissingTypeHints
class TestData:

    @pytest.fixture(scope='function')
    def test_client(self):
        sync.cleanup_all()
        from fractalis import app
        app.testing = True
        with app.test_client() as test_client:
            yield test_client
            sync.cleanup_all()

    @staticmethod
    def small_load(fail=False):
        return {
            'handler': 'test',
            'server': 'localhost:1234',
            'auth': {'token': 'fail' if fail else '7746391376142672192764'},
            'descriptors': [{
                'data_type': 'default',
                'concept': str(uuid4())
            }]
        }

    @staticmethod
    def big_load(fail=False):
        return {
            'handler': 'test',
            'server': 'localhost:1234',
            'auth': {'token': 'fail' if fail else '7746391376142672192764'},
            'descriptors': [
                {
                    'data_type': 'default',
                    'concept': str(uuid4())
                },
                {
                    'data_type': 'default',
                    'concept': str(uuid4())
                },
                {
                    'data_type': 'default',
                    'concept': str(uuid4())
                }
            ]
        }

    @pytest.fixture(scope='function', params=['small', 'big'])
    def payload(self, request):
        def _payload():
            load = self.small_load() if request.param == 'small' \
                else self.big_load()
            return {'size': len(load['descriptors']),
                    'serialized': flask.json.dumps(load)}
        return _payload

    @pytest.fixture(scope='function', params=['small', 'big'])
    def faiload(self, request):
        load = self.small_load(True) if request.param == 'small' \
            else self.big_load(True)
        return {'size': len(load['descriptors']),
                'serialized': flask.json.dumps(load)}

    @pytest.fixture(scope='function', params=[
        {
            'handler': '',
            'server': 'localhost',
            'auth': '{"''tok"n'"" '12345678"90',
            'descriptors': '[{"data_type": "default", "concept": "GSE1234"}]'
        },
        {
            'handler': 'test',
            'server': '',
            'auth': '{"token": "1234567890"}',
            'descriptors': '[{"data_type": "default", "concept": "GSE1234"}]'
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '',
            'descriptors': '[{"data_type": "default", "concept": "GSE1234"}]'
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '{"token": "1234567890"}',
            'descriptors': ''
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '{"token": "1234567890"}',
            'descriptors': '[{"data_type": "default", "concept": "GSE1234"}]'
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '{"token": "1234567890"}',
            'descriptors': '[{"concept": "GSE1234"}]'
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '{"token": "1234567890"}',
            'descriptors': '[{"data_type": "default"}]'
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '{"token": "1234567890"}',
            'descriptors': '[{"data_type": "", "concept": "GSE1234"}]'
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '{"token": "234567890"}',
            'descriptors': '[]'
        },
        {
            'handler': 'test',
            'server': 'localhost',
            'auth': '{}',
            'descriptors': '[{"data_type": "default", "concept": "GSE1234"}]'
        }
    ])
    def bad_post(self, test_client, request):
        return lambda: test_client.post('/data', data=flask.json.dumps({
            'handler': request.param['handler'],
            'server': request.param['server'],
            'auth': request.param['auth'],
            'descriptors': request.param['descriptors']
        }))

    def test_400_on_invalid_payload(self, bad_post):
        assert bad_post().status_code == 400

    def test_valid_response_on_post(self, test_client, payload):
        data = payload()
        rv = test_client.post('/data', data=data['serialized'])
        assert rv.status_code == 201
        body = flask.json.loads(rv.get_data())
        assert not body

    def test_valid_redis_before_loaded_on_post(self, test_client, payload):
        data = payload()
        test_client.post('/data', data=data['serialized'])
        keys = redis.keys('data:*')
        assert len(keys) == data['size']
        for key in keys:
            value = redis.get(key)
            data_state = json.loads(value)
            assert 'file_path' in data_state
            assert 'label' in data_state
            assert 'data_type' in data_state
            assert 'meta' in data_state

    def test_valid_redis_after_loaded_on_post(self, test_client, payload):
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        keys = redis.keys('data:*')
        assert len(keys) == data['size']
        for key in keys:
            value = redis.get(key)
            data_state = json.loads(value)
            assert 'file_path' in data_state
            assert 'label' in data_state
            assert 'data_type' in data_state
            assert 'meta' in data_state

    def test_valid_filesystem_before_loaded_on_post(
            self, test_client, payload):
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        data = payload()
        test_client.post('/data', data=data['serialized'])
        if os.path.exists(data_dir):
            assert len(os.listdir(data_dir)) == 0
        keys = redis.keys('data:*')
        for key in keys:
            value = redis.get(key)
            data_state = json.loads(value)
            assert not os.path.exists(data_state['file_path'])

    def test_valid_filesystem_after_loaded_on_post(
            self, test_client, payload):
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        assert len(os.listdir(data_dir)) == data['size']
        for f in os.listdir(data_dir):
            assert UUID(f)
        keys = redis.keys('data:*')
        for key in keys:
            value = redis.get(key)
            data_state = json.loads(value)
            assert os.path.exists(data_state['file_path'])

    def test_valid_session_on_post(self, test_client, payload):
        data = payload()
        test_client.post('/data', data=data['serialized'])
        with test_client.session_transaction() as sess:
            assert len(sess['data_tasks']) == data['size']

    def test_session_matched_redis_in_post_big_payload(
            self, test_client, payload):
        data = payload()
        test_client.post('/data', data=data['serialized'])
        with test_client.session_transaction() as sess:
            for task_id in sess['data_tasks']:
                assert redis.exists('data:{}'.format(task_id))

    def test_many_post_and_valid_state(self, test_client, payload):
        requests = 5
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        size = 0
        for i in range(requests):
            data = payload()
            size += data['size']
            rv = test_client.post('/data?wait=1', data=data['serialized'])
            assert rv.status_code == 201
        assert len(os.listdir(data_dir)) == size
        assert len(redis.keys('data:*')) == size

    def test_valid_response_before_loaded_on_get(self, test_client, payload):
        data = payload()
        test_client.post('/data', data=data['serialized'])
        rv = test_client.get('/data')
        assert rv.status_code == 200
        body = flask.json.loads(rv.get_data())
        for data_state in body['data_states']:
            assert 'file_path' not in data_state
            assert 'meta' not in data_state
            assert data_state['etl_state'] == 'SUBMITTED'
            assert not data_state['etl_message']
            assert data_state['data_type'] == 'mock'
            assert 'task_id' in data_state

    def test_valid_response_after_loaded_on_get(self, test_client, payload):
        data = payload()
        test_client.post('/data', data=data['serialized'])
        rv = test_client.get('/data?wait=1')
        assert rv.status_code == 200
        body = flask.json.loads(rv.get_data())
        for data_state in body['data_states']:
            assert 'file_path' not in data_state
            assert data_state['etl_state'] == 'SUCCESS'
            assert not data_state['etl_message']
            assert data_state['data_type'] == 'mock'
            assert 'task_id' in data_state

    def test_discard_expired_states(self, test_client):
        data_state = {
            'a': 'b',
            'file_path': '',
            'meta': ''
        }
        redis.set(name='data:456', value=json.dumps(data_state))
        with test_client.session_transaction() as sess:
            sess['data_tasks'] = ['123', '456']
        rv = test_client.get('/data?wait=1')
        body = flask.json.loads(rv.get_data())
        assert rv.status_code == 200, body
        assert len(body['data_states']) == 1
        assert body['data_states'][0]['a'] == 'b'
        with test_client.session_transaction() as sess:
            sess['data_tasks'] = ['456']

    def test_valid_response_if_failing_on_get(self, test_client, faiload):
        test_client.post('/data', data=faiload['serialized'])
        rv = test_client.get('/data?wait=1')
        assert rv.status_code == 200
        body = flask.json.loads(rv.get_data())
        for data_state in body['data_states']:
            assert 'file_path' not in data_state
            assert data_state['etl_state'] == 'FAILURE'
            assert data_state['etl_message']
            assert data_state['data_type'] == 'mock'
            assert 'task_id' in data_state

    def test_valid_state_for_finished_etl_on_delete(
            self, test_client, payload):
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        for key in redis.keys('data:*'):
            value = redis.get(key)
            data_state = json.loads(value)
            os.path.exists(data_state['file_path'])
            test_client.delete('/data/{}?wait=1'.format(data_state['task_id']))
            assert not redis.exists(key)
            assert not os.path.exists(data_state['file_path'])
            with test_client.session_transaction() as sess:
                assert data_state['task_id'] not in sess['data_tasks']

    def test_valid_state_for_running_etl_on_delete(self, test_client, payload):
        data = payload()
        test_client.post('/data', data=data['serialized'])
        for key in redis.keys('data:*'):
            value = redis.get(key)
            data_state = json.loads(value)
            assert not os.path.exists(data_state['file_path'])
            test_client.delete('/data/{}?wait=1'.format(data_state['task_id']))
            assert not redis.exists(key)
            assert not os.path.exists(data_state['file_path'])
            with test_client.session_transaction() as sess:
                assert data_state['task_id'] not in sess['data_tasks']

    def test_valid_state_for_failed_etl_on_delete(self, test_client, faiload):
        test_client.post('/data?wait=1', data=faiload['serialized'])
        for key in redis.keys('data:*'):
            value = redis.get(key)
            data_state = json.loads(value)
            assert not os.path.exists(data_state['file_path'])
            test_client.delete('/data/{}?wait=1'.format(data_state['task_id']))
            assert not redis.exists(key)
            assert not os.path.exists(data_state['file_path'])
            with test_client.session_transaction() as sess:
                assert data_state['task_id'] not in sess['data_tasks']

    def test_403_if_no_auth_on_delete(self, test_client, payload):
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        with test_client.session_transaction() as sess:
            sess['data_tasks'] = []
        for key in redis.keys('data:*'):
            value = redis.get(key)
            data_state = json.loads(value)
            assert os.path.exists(data_state['file_path'])
            rv = test_client.delete('/data/{}?wait=1'
                                    .format(data_state['task_id']))
            body = flask.json.loads(rv.get_data())
            assert rv.status_code == 403
            assert 'Refusing access.' in body['error']
            assert redis.exists(key)
            assert os.path.exists(data_state['file_path'])

    def test_valid_state_for_finished_etl_on_delete_all(
            self, test_client, payload):
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        test_client.delete('/data?wait=1')
        assert not redis.keys('data:*')
        assert len(os.listdir(data_dir)) == 0
        with test_client.session_transaction() as sess:
            assert not sess['data_tasks']

    def test_encryption_works(self, test_client, payload):
        app.config['FRACTALIS_ENCRYPT_CACHE'] = True
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        keys = redis.keys('data:*')
        for key in keys:
            value = redis.get(key)
            data_state = json.loads(value)
            file_path = data_state['file_path']
            with pytest.raises(UnicodeDecodeError):
                open(file_path, 'r').readlines()
        app.config['FRACTALIS_ENCRYPT_CACHE'] = False

    def test_valid_response_before_loaded_on_meta(self, test_client, payload):
        data = payload()
        test_client.post('/data', data=data['serialized'])
        for key in redis.keys('data:*'):
            value = redis.get(key)
            data_state = json.loads(value)
            rv = test_client.get('/data/meta/{}'.format(data_state['task_id']))
            body = flask.json.loads(rv.get_data())
            assert rv.status_code == 200
            assert 'features' not in body['meta']

    def test_valid_response_after_loaded_on_meta(self, test_client, payload):
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        for key in redis.keys('data:*'):
            value = redis.get(key)
            data_state = json.loads(value)
            rv = test_client.get('/data/meta/{}?wait=1'
                                 .format(data_state['task_id']))
            body = flask.json.loads(rv.get_data())
            assert rv.status_code == 200
            assert 'features' in body['meta']

    def test_403_if_no_auth_on_get_meta(self, test_client, payload):
        data = payload()
        test_client.post('/data?wait=1', data=data['serialized'])
        with test_client.session_transaction() as sess:
            sess['data_tasks'] = []
        for key in redis.keys('data:*'):
            value = redis.get(key)
            data_state = json.loads(value)
            rv = test_client.get('/data/meta/{}?wait=1'
                                 .format(data_state['task_id']))
            body = flask.json.loads(rv.get_data())
            assert rv.status_code == 403
            assert 'Refusing access.' in body['error']
            assert redis.exists(key)
            assert os.path.exists(data_state['file_path'])



"""This module provides shared functionality to the transmart ETLs."""

import logging
import pandas as pd
from urllib.parse import unquote_plus

import requests

from fractalis.data.etl import ETL

logger = logging.getLogger(__name__)

NUMERICAL_FIELD = 'numericValue'
CATEGORICAL_FIELD = 'stringValue'


def extract_data(server: str, descriptor: dict, token: str) -> dict:
    """Extract data from transmart.
    :param server: The target server host.
    :param descriptor: Dict describing the data to download.
    :param token: The token used for authentication.
    """
    params = dict(
        constraint=descriptor['constraint'],
        type='clinical'
    )
    if descriptor['data_type'] == 'numerical_array':
        params['type'] = 'autodetect'
        params['projection'] = 'log_intensity'

        if 'biomarker_constraint' in descriptor:
            params['biomarker_constraint'] = descriptor['biomarker_constraint']

    r = requests.get(url='{}/v2/observations'.format(server),
                     params=params,
                     headers={
                         'Accept': 'application/json',
                         'Authorization': 'Bearer {}'.format(token)
                     },
                     timeout=2000)

    logger.info('URL called: {}'.format(
        unquote_plus(r.url))
    )
    if r.status_code != 200:
        error = "Target server responded with status code {}. Message: {}.".\
            format(r.status_code, r.json())
        logger.error(error)
        raise ValueError(error)

    try:
        return r.json()
    except Exception as e:
        logger.exception(e)
        raise ValueError("Got unexpected data format.")


def get_dimension_index(obs, dimension):
    return list(obs['dimensionElements'].keys()).index(dimension)


def get_dimension_element(obs, dimension, index):
    return obs['dimensionElements'][dimension][index]


def transform_clinical(raw_data: dict, value_field: str) -> pd.DataFrame:
    patient_idx = get_dimension_index(raw_data, 'patient')
    rows = []

    for entry in raw_data['cells']:
        patient_element = entry['dimensionIndexes'][patient_idx]
        patient = get_dimension_element(raw_data, 'patient', patient_element)

        rows.append([
            patient['inTrialId'],
            entry[value_field]
        ])

    df = pd.DataFrame(rows, columns=['id', 'value'])
    feature = df.columns[1]
    df.insert(1, 'feature', feature)
    return df


def transform_highdim(raw_data: dict):
    sample_idx = get_dimension_index(raw_data, 'assay')
    feature_idx = get_dimension_index(raw_data, 'biomarker')
    rows = []

    for entry in raw_data['cells']:
        sample_element = entry['dimensionIndexes'][sample_idx]
        sample = get_dimension_element(raw_data, 'assay', sample_element)

        feature_element = entry['dimensionIndexes'][feature_idx]
        feature = get_dimension_element(raw_data, 'biomarker', feature_element)

        rows.append([
            sample['sampleCode'],
            entry[NUMERICAL_FIELD],
            feature['label']
        ])

    df = pd.DataFrame(rows, columns=['id', 'value', 'feature'])
    return df


def create_etl_type(name_, produces_, field_name):
    """
    Create a ETL task class based on a specific input type.

    :param name_: task name for registry.
    :param produces_: output type.
    :param field_name: name of cell in observation (numericValue, stringValue)
    :return: ETL task class
    """
    class TransmartETL(ETL):

        name = name_
        produces = produces_

        @staticmethod
        def can_handle(handler: str, descriptor: dict) -> bool:
            return handler == 'transmart' and descriptor['data_type'] == produces_

        def extract(self, server: str, token: str, descriptor: dict) -> dict:
            return extract_data(server=server, descriptor=descriptor, token=token)

        def transform(self, raw_data: dict, descriptor: dict) -> pd.DataFrame:
            if self.produces in ('numerical', 'categorical'):
                return transform_clinical(raw_data, field_name)
            if self.produces == 'numerical_array':
                return transform_highdim(raw_data)

    return TransmartETL



"""Module containing analysis code for heatmap analytics."""

from typing import List, TypeVar
from functools import reduce
import logging

import pandas as pd

from fractalis.analytics.task import AnalyticTask
from fractalis.analytics.tasks.shared import utils, array_stats


T = TypeVar('T')
logger = logging.getLogger(__name__)


class HeatmapTask(AnalyticTask):
    """Heatmap Analysis Task implementing AnalyticsTask. This class is a
    submittable celery task."""

    name = 'compute-heatmap'

    def main(self, numerical_arrays: List[pd.DataFrame],
             numericals: List[pd.DataFrame],
             categoricals: List[pd.DataFrame],
             ranking_method: str,
             params: dict,
             id_filter: List[T],
             max_rows: int,
             subsets: List[List[T]]) -> dict:
        # merge input data into single df
        df = reduce(lambda a, b: a.append(b), numerical_arrays)
        if not subsets:
            # empty subsets equals all samples in one subset
            subsets = [df['id'].unique().tolist()]
        else:
            # if subsets are defined we drop the rows that are not part of one
            flattened_subsets = [x for subset in subsets for x in subset]
            df = df[df['id'].isin(flattened_subsets)]
        # apply id filter
        if id_filter:
            df = df[df['id'].isin(id_filter)]
        # drop subset ids that are not in the df
        subsets = utils.drop_unused_subset_ids(df=df, subsets=subsets)
        # make sure the input data are still valid after the pre-processing
        if df.shape[0] < 1:
            error = "Either the input data set is too small or " \
                    "the subset sample ids do not match the data."
            logger.error(error)
            raise ValueError(error)

        # make matrix of input data
        df = df.pivot(index='feature', columns='id', values='value')

        # create z-score matrix used for visualising the heatmap
        z_df = [(df.iloc[i] - df.iloc[i].mean()) / df.iloc[i].std(ddof=0)
                for i in range(df.shape[0])]
        z_df = pd.DataFrame(z_df, columns=df.columns, index=df.index)

        method = 'limma'
        if ranking_method in ['mean', 'median', 'variance']:
            method = ranking_method
        # compute statistic for ranking
        stats = array_stats.get_stats(df=df, subsets=subsets,
                                      params=params,
                                      ranking_method=method)

        # sort by ranking_value
        self.sort(df, stats[ranking_method], ranking_method)
        self.sort(z_df, stats[ranking_method], ranking_method)
        self.sort(stats, stats[ranking_method], ranking_method)

        # discard rows according to max_rows
        df = df[:max_rows]
        z_df = z_df[:max_rows]
        stats = stats[:max_rows]

        # prepare output for front-end
        df['feature'] = df.index
        z_df['feature'] = z_df.index
        df = pd.melt(df, id_vars='feature', var_name='id')
        z_df = pd.melt(z_df, id_vars='feature', var_name='id')
        df = df.merge(z_df, on=['id', 'feature'])
        df.rename(columns={'value_x': 'value', 'value_y': 'zscore'},
                  inplace=True)
        df = utils.apply_subsets(df, subsets)

        return {
            'data': df.to_dict(orient='list'),
            'stats': stats.to_dict(orient='list')
        }

    @staticmethod
    def sort(df, order, method):
        order = order.tolist()
        if method == 'P.Value' or method == 'adj.P.Val':
            order = [1 - x for x in order]
        elif method == 'logFC' or method == 't':
            order = [abs(x) for x in order]
        df['sort_value'] = order
        df.sort_values('sort_value', ascending=False, inplace=True)
        df.drop('sort_value', axis=1, inplace=True)



"""The /data controller. Please refer to doc/api for more information."""

import json
import logging
from typing import Tuple, Union

from flask import Blueprint, session, request, jsonify, Response

from fractalis.data.etlhandler import ETLHandler
from fractalis.data.schema import create_data_schema
from fractalis.validator import validate_json, validate_schema
from fractalis import celery, redis
from fractalis.sync import remove_data


data_blueprint = Blueprint('data_blueprint', __name__)
logger = logging.getLogger(__name__)


@data_blueprint.route('', methods=['POST'])
@validate_json
@validate_schema(create_data_schema)
def create_data_task() -> Tuple[Response, int]:
    """Submit new ETL tasks based on the payload of the request body.
    See doc/api/ for more information.
    :return: Empty response. Everything important is stored in the session.
    """
    logger.debug("Received POST request on /data.")
    wait = request.args.get('wait') == '1'
    payload = request.get_json(force=True)
    etl_handler = ETLHandler.factory(handler=payload['handler'],
                                     server=payload['server'],
                                     auth=payload['auth'])
    task_ids = etl_handler.handle(descriptors=payload['descriptors'],
                                  data_tasks=session['data_tasks'],
                                  use_existing=False,
                                  wait=wait)
    session['data_tasks'] += task_ids
    session['data_tasks'] = list(set(session['data_tasks']))
    logger.debug("Tasks successfully submitted. Sending response.")
    return jsonify(''), 201


def get_data_state_for_task_id(task_id: str, wait: bool) -> Union[dict, None]:
    """Return data state associated with task id.
    :param task_id: The id associated with the ETL task.
    :param wait: If true and ETL is still running wait for it.
    :return: Data state that has been stored in Redis.
    """
    async_result = celery.AsyncResult(task_id)
    if wait and async_result.state == 'SUBMITTED':
        logger.debug("'wait' was set. Waiting for tasks to finish ...")
        async_result.get(propagate=False)
    value = redis.get('data:{}'.format(task_id))
    if not value:
        return None
    data_state = json.loads(value)
    # add additional information to data_state
    result = async_result.result
    if isinstance(result, Exception):  # Exception -> str
        result = "{}: {}".format(type(result).__name__, str(result))
    data_state['etl_message'] = result
    data_state['etl_state'] = async_result.state
    return data_state


@data_blueprint.route('', methods=['GET'])
def get_all_data() -> Tuple[Response, int]:
    """Get information for all tasks that have been submitted in the lifetime
    of the current session.
    See doc/api/ for more information.
    :return: Information associated with each submitted task
    """
    logger.debug("Received GET request on /data.")
    wait = request.args.get('wait') == '1'
    data_states = []
    existing_data_tasks = []
    for task_id in session['data_tasks']:
        data_state = get_data_state_for_task_id(task_id, wait)
        if data_state is None:
            warning = "Data state with task_id '{}' expired. " \
                      "Discarding...".format(task_id)
            logger.warning(warning)
            continue
        # remove internal information from response
        del data_state['file_path']
        del data_state['meta']
        # add additional information to response
        data_states.append(data_state)
        existing_data_tasks.append(task_id)
    session['data_tasks'] = existing_data_tasks
    logger.debug("Data states collected. Sending response.")
    return jsonify({'data_states': data_states}), 200


@data_blueprint.route('/<string:task_id>', methods=['DELETE'])
def delete_data(task_id: str) -> Tuple[Response, int]:
    """Remove all traces of the data associated with the given task id.
    :param task_id: The id associated with the data
    See doc/api/ for more information.
    :return: Empty response.
    """
    logger.debug("Received DELETE request on /data/task_id.")
    wait = request.args.get('wait') == '1'
    if task_id not in session['data_tasks']:
        error = "Task ID '{}' not found in session. " \
                "Refusing access.".format(task_id)
        logger.warning(error)
        return jsonify({'error': error}), 403
    session['data_tasks'].remove(task_id)
    # possibly dangerous: http://stackoverflow.com/a/29627549
    celery.control.revoke(task_id, terminate=True, signal='SIGUSR1', wait=wait)
    remove_data(task_id=task_id)
    logger.debug("Successfully removed data from session. Sending response.")
    return jsonify(''), 200


@data_blueprint.route('', methods=['DELETE'])
def delete_all_data() -> Tuple[Response, int]:
    """Remove all traces of all data associated with this session.
    :return: Empty response.
    """
    logger.debug("Received DELETE request on /data.")
    wait = request.args.get('wait') == '1'
    for task_id in session['data_tasks']:
        remove_data(task_id=task_id)
        # possibly dangerous: http://stackoverflow.com/a/29627549
        celery.control.revoke(task_id, terminate=True,
                              signal='SIGUSR1', wait=wait)
    session['data_tasks'] = []
    logger.debug("Successfully removed all data from session. "
                 "Sending response.")
    return jsonify(''), 200


@data_blueprint.route('/meta/<string:task_id>', methods=['GET'])
def get_meta_information(task_id: str) -> Tuple[Response, int]:
    """Get meta information for given task id.
    :return: meta information object stored in redis.
    """
    logger.debug("Received GET request on /data/meta/task_id.")
    wait = request.args.get('wait') == '1'
    if task_id not in session['data_tasks']:
        error = "Task ID '{}' not found in session. " \
                "Refusing access.".format(task_id)
        logger.warning(error)
        return jsonify({'error': error}), 403
    data_state = get_data_state_for_task_id(task_id, wait)
    if data_state is None:
        error = "Could not find redis entry for this task id '{}'. " \
                "The entry probably expired.".format(task_id)
        logger.error(error)
        return jsonify({'error': error}), 404
    logger.debug("Successfully gather meta information. Sending response.")
    return jsonify({'meta': data_state['meta']}), 200



"""This module contains code that is shared between the different ETLs."""

import logging
from typing import List

import pandas as pd
import requests


logger = logging.getLogger(__name__)


def make_cookie(token: str) -> dict:
    return {'PLAY2AUTH_SESS_ID': token}


def get_field(server: str, data_set: str,
              cookie: dict, projection: str) -> List[dict]:
    r = requests.get(url='{}/dataSets/records/findCustom'.format(server),
                     headers={'Accept': 'application/json'},
                     params={
                         'dataSet': data_set,
                         'projection': ['_id', projection],
                         'filterOrId': '[{{"fieldName":"{}","conditionType":"!=","value":""}}]'.format(projection)  # noqa: 501
                     },
                     cookies=cookie,
                     timeout=60)
    if r.status_code != 200:
        error = "Target server responded with " \
                "status code {}.".format(r.status_code)
        logger.error(error)
        raise ValueError(error)
    try:
        field_data = r.json()
    except Exception as e:
        logger.exception(e)
        raise TypeError("Unexpected data format. "
                        "Possible authentication error.")
    return field_data


def prepare_ids(data: List[dict]) -> List[dict]:
    new_data = []
    for row in data:
        id = row['_id']['$oid']
        del row['_id']
        row['id'] = id
        new_data.append(row)
    return new_data


def name_to_label(data: List[dict], descriptor: dict) -> List[dict]:
    try:
        label = descriptor['dictionary']['label']
    except (KeyError, TypeError):
        return data
    for row in data:
        if descriptor['dictionary']['name'] in row:
            value = row[descriptor['dictionary']['name']]
            del row[descriptor['dictionary']['name']]
            row[label] = value
    return data


def make_data_frame(data: List[dict]) -> pd.DataFrame:
    df = pd.DataFrame(data)
    df = pd.melt(df, id_vars='id', var_name='feature')
    return df



"""This module provides statistics for mRNA and miRNA data."""

from copy import deepcopy
from typing import List, TypeVar
from collections import OrderedDict
import logging

import pandas as pd
import numpy as np
from rpy2 import robjects as robj
from rpy2.robjects import r, pandas2ri
from rpy2.robjects.packages import importr


T = TypeVar('T')
importr('limma')
importr('DESeq2')
pandas2ri.activate()
logger = logging.getLogger(__name__)


def get_stats(df: pd.DataFrame, subsets: List[List[T]],
              params: dict, ranking_method: str) -> pd.DataFrame:
    if ranking_method == 'mean':
        stats = get_mean_stats(df)
    elif ranking_method == 'median':
        stats = get_median_stats(df)
    elif ranking_method == 'variance':
        stats = get_variance_stats(df)
    elif ranking_method == 'limma':
        stats = get_limma_stats(df, subsets)
    elif ranking_method == 'DESeq2':
        stats = get_deseq2_stats(df, subsets, **params)
    else:
        error = "Static method unknown: {}".format(ranking_method)
        logger.exception(error)
        raise NotImplementedError(error)
    return stats


def get_mean_stats(df: pd.DataFrame) -> pd.DataFrame:
    means = np.mean(df, axis=1)
    stats = pd.DataFrame(means, columns=['mean'])
    stats['feature'] = df.index
    return stats


def get_median_stats(df: pd.DataFrame) -> pd.DataFrame:
    medians = np.median(df, axis=1)
    stats = pd.DataFrame(medians, columns=['median'])
    stats['feature'] = df.index
    return stats


def get_variance_stats(df: pd.DataFrame) -> pd.DataFrame:
    variances = np.var(df, axis=1)
    stats = pd.DataFrame(variances, columns=['variance'])
    stats['feature'] = df.index
    return stats


def get_limma_stats(df: pd.DataFrame, subsets: List[List[T]]) -> pd.DataFrame:
    """Use the R bioconductor package 'limma' to perform a differential
    gene expression analysis on the given data frame.
    :param df: Matrix of measurements where each column represents a sample
    and each row a gene/probe.
    :param subsets: Groups to compare with each other.
    :return: Results of limma analysis. More than 2 subsets will result in
    a different structured result data frame. See ?topTableF in R.
    """
    logger.debug("Computing limma stats")
    # prepare the df in case an id exists in more than one subset
    if len(subsets) < 2:
        error = "Limma analysis requires at least " \
                "two non-empty groups for comparison."
        logger.error(error)
        raise ValueError(error)
    if df.shape[0] < 1 or df.shape[1] < 2:
        error = "Limma analysis requires a " \
                "data frame with dimension 1x2 or more."
        logger.error(error)
        raise ValueError(error)

    flattened_subsets = [x for subset in subsets for x in subset]
    df = df[flattened_subsets]
    ids = list(df)
    features = df.index

    # creating the design vector according to the subsets
    design_vector = [''] * len(ids)
    subsets_copy = deepcopy(subsets)
    for i, sample_id in enumerate(ids):
        for j, subset in enumerate(subsets_copy):
            try:
                subset.index(sample_id)  # raises an Exception if not found
                subset.remove(sample_id)
                design_vector[i] = str(j + 1)
                break
            except ValueError:
                assert j != len(subsets_copy) - 1

    assert '' not in design_vector

    # create group names
    groups = ['group{}'.format(i + 1) for i in list(range(len(subsets)))]

    # create a string for each pairwise comparison of the groups
    comparisons = []
    for i in reversed(range(len(subsets))):
        for j in range(i):
            comparisons.append('group{}-group{}'.format(i+1, j+1))

    # fitting according to limma doc Chapter 8: Linear Models Overview
    r_form = robj.Formula('~ 0+factor(c({}))'.format(','.join(design_vector)))
    r_design = r['model.matrix'](r_form)
    r_design.colnames = robj.StrVector(groups)
    r_data = pandas2ri.py2ri(df)
    # py2ri is stupid and makes too many assumptions.
    # These two lines restore the column order
    r_data.colnames = list(OrderedDict.fromkeys(ids))
    r_data = r_data.rx(robj.StrVector(ids))

    r_fit = r['lmFit'](r_data, r_design)
    r_contrast_matrix = r['makeContrasts'](*comparisons, levels=r_design)
    r_fit_2 = r['contrasts.fit'](r_fit, r_contrast_matrix)
    r_fit_2 = r['eBayes'](r_fit_2)
    r_results = r['topTable'](r_fit_2, number=float('inf'),
                              sort='none', genelist=features)
    results = pandas2ri.ri2py(r_results)
    # let's give the gene list column an appropriate name
    colnames = results.columns.values
    colnames[0] = 'feature'
    results.columns = colnames

    return results


# FIXME: Add more parameters (e.g. for filtering low count rows)
def get_deseq2_stats(df: pd.DataFrame,
                     subsets: List[List[T]],
                     min_total_row_count: int = 0) -> pd.DataFrame:
    """Use the R bioconductor package 'limma' to perform a differential
    expression analysis of count like data (e.g. miRNA). See package
    documentation for more details.
    :param df: Matrix of counts, where each column is a sample and each row
    a feature.
    :param subsets: The two subsets to compare with each other.
    :param min_total_row_count: Drop rows that have in total less than than
        min_total_row_count reads
    :return: Results of the analysis in form of a Dataframe (p, logFC, ...)
    """
    logger.debug("Computing deseq2 stats")
    if len(subsets) != 2:
        error = "This method currently only supports exactly two " \
                "subsets as this is the most common use case. Support " \
                "for more subsets will be added later."
        logger.exception(error)
        raise ValueError(error)
    # flatten subset
    flattened_subsets = [x for subset in subsets for x in subset]
    # discard columns that are not in a subset
    df = df[flattened_subsets]
    # filter rows with too few reads
    total_row_counts = df.sum(axis=1)
    keep = total_row_counts[total_row_counts >= min_total_row_count].index
    df = df.loc[keep]
    # pandas df -> R df
    r_count_data = pandas2ri.py2ri(df)
    # py2ri is stupid and makes too many assumptions.
    # These two lines restore the column order
    r_count_data.colnames = list(OrderedDict.fromkeys(flattened_subsets))
    r_count_data = r_count_data.rx(robj.StrVector(flattened_subsets))

    # see package documentation
    condition = ['s{}'.format(i) for i, subset in enumerate(subsets)
                 for _ in subset]
    r_condition = robj.FactorVector(robj.StrVector(condition))
    r_col_data = r['DataFrame'](condition=r_condition)
    r_design = robj.Formula('~ condition')
    r_design.environment['condition'] = r_condition
    r_dds = r['DESeqDataSetFromMatrix'](r_count_data, r_col_data, r_design)
    r_dds = r['DESeq'](r_dds, parallel=True)
    r_res = r['results'](r_dds)

    # R result table to Python pandas
    r_res = r['as.data.frame'](r_res)
    results = pandas2ri.ri2py(r_res)
    results.insert(0, 'feature', list(r['row.names'](r_res)))
    return results



from time import sleep

import requests

from fractalis import app


def submit_query(query: str, server: str, token: str) -> int:
    r = requests.post(
        url='{}/queryService/runQuery'.format(server),
        data=query,
        headers={
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {}'.format(token)
        },
        verify=app.config['ETL_VERIFY_SSL_CERT']
    )
    r.raise_for_status()
    result_id = r.json()['resultId']
    return result_id


def wait_for_completion(result_id: int, server, token):
    def _check_status():
        return requests.get(
            url='{}/resultService/resultStatus/{}'.format(
                server, result_id),
            headers={
                'Content-Type': 'application/json',
                'Authorization': 'Bearer {}'.format(token)
            },
            verify=app.config['ETL_VERIFY_SSL_CERT']
        ).json()
    while _check_status()['status'] == 'RUNNING':
        sleep(1)


def get_data(result_id, server, token):
    r = requests.get(
        url='{}/resultService/result/{}/CSV'.format(
            server, result_id),
        headers={
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {}'.format(token)
        },
        verify=app.config['ETL_VERIFY_SSL_CERT']
    )
    r.raise_for_status()
    return r.text



"""This module provides tests for the etlhandler module."""

import os
from pathlib import Path

import pytest

from fractalis import celery, app
from fractalis.data.etlhandler import ETLHandler


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints,PyPep8Naming
class TestETLHandler:

    @pytest.fixture(scope='function')
    def redis(self):
        from fractalis import redis, sync
        yield redis
        sync.cleanup_all()

    etlhandler = ETLHandler.factory(handler='test', server='localfoo', auth={})

    def test_descriptor_to_hash_produces_unique_hash(self, redis):
        hash_1 = self.etlhandler.descriptor_to_hash(descriptor={'a': 1})
        hash_2 = self.etlhandler.descriptor_to_hash(descriptor={'': ''})
        hash_3 = self.etlhandler.descriptor_to_hash(descriptor={'a': 1})
        self.etlhandler._server = 'localbar'
        hash_4 = self.etlhandler.descriptor_to_hash(descriptor={'a': 1})
        assert isinstance(hash_1, int)
        assert isinstance(hash_4, int)
        assert hash_1 == hash_3
        assert hash_1 != hash_2
        assert hash_1 != hash_4

    def test_find_duplicates_finds_duplicates_by_hash(self, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        duplicates = self.etlhandler.find_duplicates(data_tasks=['123'],
                                                     descriptor=descriptor)
        assert len(duplicates) == 1
        assert duplicates[0] == '123'

    def test_finds_all_duplicates(self, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='456',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='789',
                                           file_path='',
                                           descriptor={'a': 5},
                                           data_type='')
        duplicates = self.etlhandler.find_duplicates(
            data_tasks=['123', '456', '789'], descriptor=descriptor)
        assert len(duplicates) == 2
        assert '123' in duplicates
        assert '456' in duplicates

    def test_find_duplicates_only_operates_on_given_list(self, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='456',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='789',
                                           file_path='',
                                           descriptor={'a': 5},
                                           data_type='')
        duplicates = self.etlhandler.find_duplicates(
            data_tasks=['456', '789'], descriptor=descriptor)
        assert len(duplicates) == 1
        assert '456' in duplicates

    def test_remove_duplicates_removes_duplicate(self, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='456',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        os.makedirs(data_dir, exist_ok=True)
        Path(os.path.join(data_dir, '456')).touch()
        assert redis.exists('data:456')
        self.etlhandler.remove_duplicates(data_tasks=['456'],
                                          descriptor=descriptor)
        assert not redis.exists('data:456')

    def test_remove_duplicates_removes_all_duplicates(self, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='456',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='789',
                                           file_path='',
                                           descriptor={'a': 1},
                                           data_type='')
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        os.makedirs(data_dir, exist_ok=True)
        Path(os.path.join(data_dir, '123')).touch()
        Path(os.path.join(data_dir, '456')).touch()
        Path(os.path.join(data_dir, '789')).touch()
        assert redis.exists('data:123')
        assert redis.exists('data:456')
        assert redis.exists('data:789')
        self.etlhandler.remove_duplicates(data_tasks=['123', '456', '789'],
                                          descriptor=descriptor)
        assert not redis.exists('data:123')
        assert not redis.exists('data:456')
        assert redis.exists('data:789')

    def test_remove_duplicates_only_operates_on_given_list(self, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='456',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        self.etlhandler.create_redis_entry(task_id='789',
                                           file_path='',
                                           descriptor={'a': 1},
                                           data_type='')
        data_dir = os.path.join(app.config['FRACTALIS_TMP_DIR'], 'data')
        os.makedirs(data_dir, exist_ok=True)
        Path(os.path.join(data_dir, '123')).touch()
        Path(os.path.join(data_dir, '456')).touch()
        Path(os.path.join(data_dir, '789')).touch()
        assert redis.exists('data:123')
        assert redis.exists('data:456')
        assert redis.exists('data:789')
        self.etlhandler.remove_duplicates(data_tasks=['123', '789'],
                                          descriptor=descriptor)
        assert not redis.exists('data:123')
        assert redis.exists('data:456')
        assert redis.exists('data:789')

    def test_find_duplicate_task_id_returns_task_id_of_SUCCESS(
            self, monkeypatch, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUCCESS'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        task_id = self.etlhandler.find_duplicate_task_id(
            data_tasks=['123'], descriptor=descriptor)
        assert task_id == '123'

    def test_find_duplicate_task_id_returns_task_id_of_SUBMITTED(
            self, monkeypatch, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUBMITTED'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        task_id = self.etlhandler.find_duplicate_task_id(
            data_tasks=['123'], descriptor=descriptor)
        assert task_id == '123'

    def test_find_duplicate_task_id_returns_None_for_FAILURE(
            self, monkeypatch, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'FAILURE'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        task_id = self.etlhandler.find_duplicate_task_id(
            data_tasks=['123'], descriptor=descriptor)
        assert task_id is None

    def test_find_duplicate_limits_search_to_data_tasks(
            self, monkeypatch, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUCCESS'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        task_id = self.etlhandler.find_duplicate_task_id(
            data_tasks=['456'], descriptor=descriptor)
        assert task_id is None

    def test_find_duplicate_task_id_returns_None_for_not_existing(
            self, monkeypatch, redis):
        descriptor = {'a': {'b': 3}, 'c': 4}

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'FAILURE'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        task_id = self.etlhandler.find_duplicate_task_id(
            data_tasks=['123'], descriptor=descriptor)
        assert task_id is None

    def test_handle_reuses_existing_task_ids_if_use_existing(
            self, monkeypatch, redis):
        descriptor = {'data_type': 'default'}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')

        class FakeAsyncResult:
            def __init__(self, *args, **kwargs):
                self.state = 'SUBMITTED'

            def get(self, *args, **kwargs):
                pass
        monkeypatch.setattr(celery, 'AsyncResult', FakeAsyncResult)
        task_ids = self.etlhandler.handle(descriptors=[descriptor],
                                          data_tasks=['123'],
                                          use_existing=True)
        assert len(task_ids) == 1
        assert task_ids[0] == '123'

    def test_handle_limits_search_to_tasks_ids(self, monkeypatch, redis):
        descriptor = {'data_type': 'default'}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        task_ids = self.etlhandler.handle(descriptors=[descriptor],
                                          data_tasks=[],
                                          use_existing=True)
        assert len(task_ids) == 1
        assert task_ids[0] != '123'

    def test_handle_removes_old_and_returns_new_if_not_use_existing(
            self, monkeypatch, redis):
        descriptor = {'data_type': 'default'}
        self.etlhandler.create_redis_entry(task_id='123',
                                           file_path='',
                                           descriptor=descriptor,
                                           data_type='')
        task_ids = self.etlhandler.handle(descriptors=[descriptor],
                                          data_tasks=['123'],
                                          use_existing=False)
        assert len(task_ids) == 1
        assert task_ids[0] != '123'

    def test_handle_removes_duplicate_of_previous_iteration(
            self, monkeypatch, redis):
        descriptor = {'data_type': 'default'}
        task_ids = self.etlhandler.handle(descriptors=[descriptor, descriptor],
                                          data_tasks=[],
                                          use_existing=False)
        assert len(task_ids) == 2
        assert task_ids[0] != task_ids[1]
        assert len(redis.keys('data:*')) == 1

    def test_handle_uses_duplicate_of_previous_iteration(
            self, monkeypatch, redis):
        descriptor = {'data_type': 'default'}
        task_ids = self.etlhandler.handle(descriptors=[descriptor, descriptor],
                                          data_tasks=[],
                                          use_existing=True)
        assert len(task_ids) == 1
        assert len(redis.keys('data:*')) == 1



"""Provides StringETL for Ada."""

from typing import List

from pandas import DataFrame

from fractalis.data.etl import ETL
from fractalis.data.etls.ada import shared


class StringETL(ETL):
    """StringETL implements support for Adas 'String' type"""

    name = 'ada_string_etl'
    produces = 'categorical'

    @staticmethod
    def can_handle(handler, descriptor):
        return handler == 'ada' and \
               descriptor['dictionary']['fieldType'] and \
               descriptor['dictionary']['fieldType'] == 'String' and not \
               descriptor['dictionary']['isArray']

    def extract(self, server: str, token: str, descriptor: dict) -> List[dict]:
        data_set = descriptor['data_set']
        projection = descriptor['dictionary']['projection']
        cookie = shared.make_cookie(token=token)
        data = shared.get_field(server=server, data_set=data_set,
                                cookie=cookie, projection=projection)
        return data

    def transform(self, raw_data: List[dict], descriptor: dict) -> DataFrame:
        data = shared.prepare_ids(raw_data)
        data = shared.name_to_label(data, descriptor)
        df = shared.make_data_frame(data)
        return df



# -*- coding: utf-8 -*-

from distutils.core import setup
from setuptools import find_packages

setup(
    name='fractalis',
    packages=find_packages(),
    author='Sascha Herzinger',
    author_email='sascha.herzinger@uni.lu',
    url='https://git-r3lab.uni.lu/Fractalis/fractalis',
    version='1.1.0',
    license='Apache2.0',
    include_package_data=True,
    python_requires='>=3.4.0',
    install_requires=[
        'Flask==0.12.2',
        'flask-cors==3.0.3',
        'Flask-Script==2.0.6',
        'flask-request-id-middleware==1.1',
        'flask-compress==1.4.0',
        'typing==3.6.2',
        'jsonschema==2.6.0',
        'celery[redis]==4.1.0',
        'kombu==4.1.0',
        'redis==2.10.6',
        'numpy==1.13.3',
        'scipy==0.19.1',
        'pandas==0.20.3',
        'scikit-learn==0.19.1',
        'lifelines==0.14.3',
        'requests==2.18.4',
        'PyYAML==3.12',
        'pycryptodomex==3.4.7',
        'rpy2==2.9.3',
        'tzlocal',
        'flake8',
        'pytest',
        'pytest-runner',
        'pytest-cov',
        'responses',
        'twine'
    ]
)



"""Module containing the Celery Task for the Correlation Analysis."""

import logging
from typing import List

import pandas as pd
import numpy as np
from scipy import stats

from fractalis.analytics.task import AnalyticTask
from fractalis.analytics.tasks.shared import utils


logger = logging.getLogger(__name__)


class CorrelationTask(AnalyticTask):
    """Correlation Analysis Task implementing AnalyticsTask. This class is a
    submittable celery task."""

    name = 'compute-correlation'

    def main(self,
             x: pd.DataFrame,
             y: pd.DataFrame,
             id_filter: List[str],
             method: str,
             subsets: List[List[str]],
             categories: List[pd.DataFrame]) -> dict:
        """Compute correlation statistics for the given parameters.
        :param x: DataFrame containing x axis values.
        :param y: DataFrame containing y axis values.
        :param id_filter: If specified use only given ids during the analysis.
        :param method: pearson, spearman or kendall.
        :param subsets: List of lists of subset ids.
        :param categories: List of DataFrames that categorise the data points.
        :return: corr. coef., p-value and other useful values.
        """
        if len(x['feature'].unique().tolist()) != 1 \
                or len(y['feature'].unique().tolist()) != 1:
            error = "Input is invalid. Please make sure that the two " \
                    "variables to compare have exactly one dimension, each."
            logger.error(error)
            raise ValueError(error)
        if method not in ['pearson', 'spearman', 'kendall']:
            raise ValueError("Unknown method '{}'".format(method))

        df = self.merge_x_y(x, y)
        x_label = list(df['feature_x'])[0]
        y_label = list(df['feature_y'])[0]
        if id_filter:
            df = df[df['id'].isin(id_filter)]
        df = utils.apply_subsets(df=df, subsets=subsets)
        df = utils.apply_categories(df=df, categories=categories)
        global_stats = self.compute_stats(df, method)
        output = global_stats
        output['method'] = method
        output['data'] = df.to_json(orient='records')
        output['x_label'] = x_label
        output['y_label'] = y_label

        return output

    @staticmethod
    def merge_x_y(x: pd.DataFrame, y: pd.DataFrame) -> pd.DataFrame:
        """Merge the x and y DataFrame and drop all rows containing NA.
        :param x: The x-axis values.
        :param y: The y-axis values.
        :return: The merged data frame.
        """
        df = x.merge(y, on=['id'], how='inner')
        df = df.dropna()
        if df.shape[0] == 0:
            raise ValueError("X and Y do not share any ids.")
        return df

    @staticmethod
    def compute_stats(df: pd.DataFrame, method: str) -> dict:
        """Compute correlation statistics for the given data and the given
        correlation method.
        :param df: The DataFrame containing our data.
        :param method: The method to use.
        :return: Several relevant statistics
        """
        df = df.drop_duplicates('id')
        df = df.dropna()
        if df.shape[0] < 2:
            return {
                'coef': float('nan'),
                'p_value': float('nan'),
                'slope': float('nan'),
                'intercept': float('nan')
            }
        if method == 'pearson':
            corr_coef, p_value = stats.pearsonr(df['value_x'], df['value_y'])
        elif method == 'spearman':
            corr_coef, p_value = stats.spearmanr(df['value_x'], df['value_y'])
        elif method == 'kendall':
            corr_coef, p_value = stats.kendalltau(df['value_x'], df['value_y'])
        else:
            raise ValueError("Unknown correlation method.")
        slope, intercept, *_ = np.polyfit(df['value_x'], df['value_y'], deg=1)
        return {
            'coef': corr_coef,
            'p_value': p_value,
            'slope': slope,
            'intercept': intercept,
        }



"""This module provides tests for the AnalyticsTask class."""

import pandas as pd

from uuid import uuid4
from fractalis.analytics.task import AnalyticTask


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class MockTask(AnalyticTask):
    @property
    def name(self) -> str:
        return ''

    def main(self, *args, **kwargs) -> dict:
        pass


# noinspection PyMissingOrEmptyDocstring,PyMissingTypeHints
class TestAnalyticsTask:

    task = MockTask()

    def test_apply_filter(self):
        df = pd.DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                          columns=['A', 'B', 'C'])
        filters1 = {'A': [1, 7], 'B': [2, 5, 8], 'C': [6, 9]}
        filters2 = {'A': []}
        df1 = self.task.apply_filters(df=df, filters=filters1)
        df2 = self.task.apply_filters(df=df, filters=filters2)
        assert df1.shape == (1, 3)
        assert df1.iloc[0].tolist() == [7, 8, 9]
        assert df2.shape == (3, 3)

    def test_parse_value(self):
        uuid = str(uuid4())
        arg1 = '${"id": 123, "filters": {"foo": [1,2]}}$'
        arg2 = '${}$'.format(uuid)
        arg3 = '${{"id": "{}", "filters": {{"foo": [1,2]}}}}$'.format(uuid)
        data_task_id, filters = self.task.parse_value(arg1)
        assert data_task_id is None
        assert 'foo' in filters
        assert filters['foo'] == [1, 2]
        data_task_id, filters = self.task.parse_value(arg2)
        assert data_task_id == uuid
        assert not filters
        data_task_id, filters = self.task.parse_value(arg3)
        assert data_task_id == uuid
        assert 'foo' in filters
        assert filters['foo'] == [1, 2]



from fractalis.data.etlhandler import ETLHandler


class TCGADemoHandler(ETLHandler):

    _handler = 'demo_tcga_coad'

    @staticmethod
    def make_label(descriptor):
        return descriptor.get('field')

    def _get_token_for_credentials(self, server: str, auth: dict) -> str:
        return 'foo'

    def _heartbeat(self):
        pass



"""Provides highdim concept ETL for tranSMART."""

from fractalis.data.etls.transmart.shared import create_etl_type, NUMERICAL_FIELD

HighdimETL = create_etl_type(
    name_='transmart_highdim_etl',
    produces_='numerical_array',
    field_name=NUMERICAL_FIELD
)
